import abc
import collections
import copy
import re
from collections.abc import Iterable

from supriya.system import SupriyaObject


class ControlInterface(SupriyaObject):

    # TODO: This should inherit from collections.Mapping

    ### CLASS VARIABLES ###

    __slots__ = ("_synth_controls", "_client")

    _bus_pattern = re.compile(r"(?P<type>c|a)(?P<id>\d+)")

    ### SPECIAL METHODS ###

    @abc.abstractmethod
    def __setitem__(self, items, values):
        raise NotImplementedError

    ### PRIVATE METHODS ###

    def _set(self, **settings):
        # TODO: Reimplement as _apply_local() on request classes
        import supriya.commands
        import supriya.realtime
        import supriya.synthdefs

        n_set_settings = {}
        n_map_settings = {}
        n_mapa_settings = {}
        for control_name, value in settings.items():
            if control_name not in self:
                continue
            control = self[control_name]
            if isinstance(value, str):
                match = self._bus_pattern.match(value)
                if match:
                    group_dict = match.groupdict()
                    if group_dict["type"] == "c":
                        calculation_rate = "control"
                    else:
                        calculation_rate = "audio"
                    value = supriya.realtime.Bus(
                        bus_group_or_index=int(group_dict["id"]),
                        calculation_rate=calculation_rate,
                    )
            if isinstance(value, (int, float)):
                n_set_settings[control_name] = float(value)
                control._set_to_number(value)
            elif isinstance(value, (supriya.realtime.Bus, supriya.realtime.BusGroup)):
                value_rate = value.calculation_rate
                if (
                    isinstance(self.client, supriya.realtime.Synth)
                    and not self.client.is_allocated
                    and control.calculation_rate == supriya.CalculationRate.SCALAR
                ):
                    control._set_to_number(int(value))
                elif value_rate == supriya.CalculationRate.CONTROL:
                    n_map_settings[control_name] = value
                    control._map_to_bus(value)
                elif value_rate == supriya.CalculationRate.AUDIO:
                    n_mapa_settings[control_name] = value
                    control._map_to_bus(value)
            elif value is None:
                n_map_settings[control_name] = -1
                control._unmap()
            else:
                try:
                    control._set_to_number(float(value))
                except TypeError:
                    raise ValueError(value)
        requests = []
        if self.client.is_allocated:
            if n_set_settings:
                request = supriya.commands.NodeSetRequest(
                    self.node_id, **n_set_settings
                )
                requests.append(request)
            if n_map_settings:
                request = supriya.commands.NodeMapToControlBusRequest(
                    self.node_id, **n_map_settings
                )
                requests.append(request)
            if n_mapa_settings:
                request = supriya.commands.NodeMapToAudioBusRequest(
                    self.node_id, **n_mapa_settings
                )
                requests.append(request)
        return tuple(requests)

    ### PUBLIC METHODS ###

    @abc.abstractmethod
    def reset(self):
        raise NotImplementedError

    ### PUBLIC PROPERTIES ###

    @property
    def client(self):
        return self._client

    @property
    def node_id(self):
        return int(self.client)


class GroupControl:

    ### CLASS VARIABLES ###

    __slots__ = ("_client", "_name")

    ### INITIALIZER ###

    def __init__(self, client=None, name=None):
        self._client = client
        self._name = str(name)

    ### SPECIAL METHODS ###

    def __repr__(self):
        class_name = type(self).__name__
        calculation_rates = sorted(
            set(
                synth.controls[self.name].calculation_rate
                for synth in self.client._synth_controls["amplitude"]
            )
        )
        return '<{}: {!r} "{}" [{}]>'.format(
            class_name,
            self.client.client,
            self.name,
            ", ".join(_.token for _ in calculation_rates),
        )

    def __str__(self):
        return self.name

    ### PRIVATE METHODS ###

    def _map_to_bus(self, bus):
        pass

    def _set_to_number(self, number):
        pass

    def _unmap(self):
        pass

    ### PUBLIC METHODS ###

    def set(self, expr):
        self._client[self.name] = expr

    ### PUBLIC PROPERTIES ###

    @property
    def client(self):
        return self._client

    @property
    def group(self):
        return self.client.client

    @property
    def name(self):
        return self._name

    @property
    def node(self):
        return self.client.client


class GroupInterface(ControlInterface):
    """
    Interface to group controls.

    ::

        >>> server = supriya.Server().boot()
        >>> group = supriya.Group().allocate(server)
        >>> group.extend(
        ...     [
        ...         supriya.Synth(synthdef=supriya.assets.synthdefs.test),
        ...         supriya.Synth(synthdef=supriya.assets.synthdefs.default),
        ...         supriya.Synth(synthdef=supriya.assets.synthdefs.default),
        ...     ]
        ... )

    ::

        >>> control = group.controls["amplitude"]

    ::

        >>> group.controls["frequency"] = 777

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_group_controls",)

    ### INITIALIZER ###

    def __init__(self, client=None):
        self._synth_controls = {}
        self._group_controls = {}
        self._client = client

    ### SPECIAL METHODS ###

    def __contains__(self, item):
        return item in self._synth_controls

    def __getitem__(self, item):
        return self._group_controls[item]

    def __iter__(self):
        return iter(sorted(self._group_controls))

    def __len__(self):
        return len(self._group_controls)

    def __repr__(self):
        class_name = type(self).__name__
        return "<{}: {!r}>".format(class_name, self.client)

    def __setitem__(self, items, values):
        import supriya.realtime

        if not isinstance(items, tuple):
            items = (items,)
        assert all(_ in self._synth_controls for _ in items)
        if not isinstance(values, tuple):
            values = (values,)
        assert len(items) == len(values)
        settings = dict(zip(items, values))
        for key, value in settings.items():
            for synth in self._synth_controls.get(key, ()):
                control = synth.controls[key]
                if isinstance(value, supriya.realtime.Bus):
                    control._map_to_bus(value)
                elif value is None:
                    control._unmap()
                else:
                    control._set_to_number(value)
        requests = self._set(**settings)
        supriya.commands.RequestBundle(contents=requests).communicate(
            server=self.client.server, sync=True
        )

    ### PUBLIC METHODS ###

    def add_controls(self, control_interface_dict):
        import supriya.realtime

        for control_name in control_interface_dict:
            if control_name not in self._synth_controls:
                self._synth_controls[control_name] = copy.copy(
                    control_interface_dict[control_name]
                )
                proxy = supriya.realtime.GroupControl(client=self, name=control_name)
                self._group_controls[control_name] = proxy
            else:
                self._synth_controls[control_name].update(
                    control_interface_dict[control_name]
                )

    def as_dict(self):
        result = {}
        for control_name, node_set in self._synth_controls.items():
            result[control_name] = copy.copy(node_set)
        return result

    def remove_controls(self, control_interface_dict):
        for control_name in control_interface_dict:
            if control_name not in self._synth_controls:
                continue
            current_nodes = self._synth_controls[control_name]
            nodes_to_remove = control_interface_dict[control_name]
            current_nodes.difference_update(nodes_to_remove)
            if not current_nodes:
                del self._synth_controls[control_name]
                del self._group_controls[control_name]

    def reset(self):
        self._synth_controls.clear()


class SynthControl:

    ### CLASS VARIABLES ###

    __slots__ = (
        "_calculation_rate",
        "_client",
        "_index",
        "_default_value",
        "_last_unmapped_value",
        "_name",
        "_range",
        "_unit",
        "_value",
        "__weakref__",
    )

    ### INITIALIZER ###

    def __init__(
        self,
        client=None,
        index=None,
        name=None,
        range_=None,
        calculation_rate=None,
        unit=None,
        value=None,
    ):
        import supriya.realtime
        import supriya.synthdefs

        self._client = client
        self._name = str(name)
        if isinstance(range_, supriya.synthdefs.Range):
            self._range = range_
        else:
            self._range = None
        self._calculation_rate = supriya.CalculationRate.from_expr(calculation_rate)
        self._unit = unit
        self._value = value
        self._default_value = value
        if not isinstance(value, supriya.realtime.Bus):
            self._last_unmapped_value = self._value
        else:
            self._last_unmapped_value = self._default_value
        if index is not None:
            index = int(index)
        self._index = index

    ### SPECIAL METHODS ###

    def __call__(self, expr):
        return self.set(expr)

    def __repr__(self):
        class_name = type(self).__name__
        return '<{}: {!r} "{}": {} [{}]>'.format(
            class_name,
            self.client.client,
            self.name,
            self.value,
            self.calculation_rate.token,
        )

    def __str__(self):
        return self.name

    ### PRIVATE METHODS ###

    def _map_to_bus(self, bus):
        import supriya.realtime

        if not isinstance(self.value, supriya.realtime.Bus):
            self._last_unmapped_value = self._value
        self._value = bus

    def _set_to_number(self, value):
        self._value = float(value)
        self._last_unmapped_value = self._value

    def _unmap(self):
        self._value = self._last_unmapped_value

    ### PUBLIC METHODS ###

    @classmethod
    def from_parameter(cls, parameter, index=0, client=None):
        import supriya.synthdefs

        assert isinstance(parameter, supriya.synthdefs.Parameter)
        name = parameter.name
        range_ = parameter.range_
        calculation_rate = supriya.CalculationRate.from_expr(parameter)
        unit = parameter.unit
        value = parameter.value
        synth_control = SynthControl(
            client=client,
            index=index,
            name=name,
            range_=range_,
            calculation_rate=calculation_rate,
            unit=unit,
            value=value,
        )
        return synth_control

    def get(self):
        return self._value

    def reset(self):
        self._value = self._default_value

    def set(self, expr):
        import supriya.commands
        import supriya.realtime
        import supriya.synthdefs

        if isinstance(expr, supriya.realtime.Bus):
            self._map_to_bus(expr)
            if expr.calculation_rate == supriya.CalculationRate.CONTROL:
                request = supriya.commands.NodeMapToControlBusRequest(
                    self.node, **{self.name: self._value}
                )
            else:
                request = supriya.commands.NodeMapToAudioBusRequest(
                    self.node, **{self.name: self._value}
                )
        elif expr is None:
            self._unmap()
            request = supriya.commands.NodeMapToControlBusRequest(
                self.node, **{self.name: -1}
            )
        else:
            self._set_to_number(expr)
            request = supriya.commands.NodeSetRequest(
                self.node, **{self.name: self._value}
            )
        if self.node.is_allocated:
            request.communicate(server=self.node.server)
        return self.get()

    ### PUBLIC PROPERTIES ###

    @property
    def calculation_rate(self):
        return self._calculation_rate

    @property
    def client(self):
        return self._client

    @property
    def default_value(self):
        return self._default_value

    @property
    def index(self):
        return self._index

    @property
    def last_unmapped_value(self):
        return self._last_unmapped_value

    @property
    def name(self):
        return self._name

    @property
    def range_(self):
        return self._range

    @property
    def node(self):
        return self.client.client

    @property
    def synth(self):
        return self.client.client

    @property
    def unit(self):
        return self._unit

    @property
    def value(self):
        return self._value


class SynthInterface(ControlInterface):

    ### CLASS VARIABLES ###

    __slots__ = ("_synthdef", "_synth_control_map")

    ### INITIALIZER ###

    def __init__(self, client=None, synthdef=None):
        import supriya.realtime
        import supriya.synthdefs

        assert isinstance(synthdef, supriya.synthdefs.SynthDef)
        self._client = client
        synth_controls = []
        synth_control_map = collections.OrderedDict()
        for index, parameter in synthdef.indexed_parameters:
            synth_control = supriya.realtime.SynthControl.from_parameter(
                parameter, client=self, index=index
            )
            synth_controls.append(synth_control)
            synth_control_map[synth_control.name] = synth_control
        self._synth_controls = tuple(synth_controls)
        self._synth_control_map = synth_control_map
        self._synthdef = synthdef or self._client.synthdef

    ### SPECIAL METHODS ###

    def __contains__(self, item):
        return item in self._synth_control_map

    def __getitem__(self, item):
        if isinstance(item, str):
            return self._synth_control_map[item]
        elif isinstance(item, Iterable):
            return tuple(self._synth_control_map[x] for x in item)
        raise ValueError

    def __iter__(self):
        return iter(sorted(self._synth_control_map))

    def __len__(self):
        return len(self._synth_control_map)

    def __repr__(self):
        """
        Get interpreter representation of synth interface.

        ::

            >>> synth = supriya.Synth()
            >>> print(repr(synth.controls))
            <SynthInterface: <- Synth: ??? default>>

        """
        class_name = type(self).__name__
        return "<{}: {!r}>".format(class_name, self.client)

    def __setitem__(self, items, values):
        import supriya.realtime

        if not isinstance(items, tuple):
            items = (items,)
        if not isinstance(values, tuple):
            values = (values,)
        assert len(items) == len(values)
        synth_controls = self.__getitem__(items)
        synth_control_names = [x.name for x in synth_controls]
        settings = dict(zip(synth_control_names, values))
        requests = self._set(**settings)
        if not self.client.is_allocated:
            return
        supriya.commands.RequestBundle(contents=requests).communicate(
            server=self.client.server, sync=True
        )

    def __str__(self):
        """
        Get string representation of synth interface.

        ::

            >>> synth = supriya.Synth()
            >>> print(str(synth.controls))
            <- Synth: ??? default>:
                (kr) amplitude: 0.1
                (kr) frequency: 440.0
                (kr) gate:      1.0
                (ir) out:       0.0
                (kr) pan:       0.5

        """
        result = []
        string = "{!r}:".format(self.client)
        result.append(string)
        maximum_length = 0
        control_names = sorted(self)
        maximum_length = max(maximum_length, max(len(_) for _ in control_names))
        maximum_length += 1
        for control_name in control_names:
            synth_control = self[control_name]
            value = str(synth_control.value)
            spacing = " " * (maximum_length - len(control_name))
            string = "    ({}) {}:{}{}".format(
                synth_control.calculation_rate.token, control_name, spacing, value
            )
            result.append(string)
        result = "\n".join(result)
        return result

    ### PRIVATE METHODS ###

    def _make_synth_new_settings(self):
        import supriya.commands
        import supriya.realtime
        import supriya.synthdefs

        audio_map = {}
        control_map = {}
        requests = []
        settings = {}
        for synth_control in self.synth_controls:
            if isinstance(synth_control.value, supriya.realtime.Bus):
                if (
                    synth_control.value.calculation_rate
                    == supriya.CalculationRate.AUDIO
                ):
                    audio_map[synth_control.name] = synth_control.value
                else:
                    control_map[synth_control.name] = synth_control.value
            elif synth_control.value != synth_control.default_value:
                settings[synth_control.name] = synth_control.value
        if audio_map:
            request = supriya.commands.NodeMapToAudioBusRequest(
                node_id=self.client, **audio_map
            )
            requests.append(request)
        if control_map:
            request = supriya.commands.NodeMapToControlBusRequest(
                node_id=self.client, **control_map
            )
            requests.append(request)
        return settings, requests

    ### PUBLIC METHODS ###

    def as_dict(self):
        result = {}
        for control_name in self:
            control = self[control_name]
            result[control.name] = set([self.client])
        return result

    def reset(self):
        for synth_control in self._synth_controls:
            synth_control.reset()

    ### PUBLIC PROPERTIES ###

    @property
    def synthdef(self):
        return self._synthdef

    @property
    def synth_controls(self):
        return self._synth_controls



from .asynchronous import AsyncClock
from .ephemera import ClockContext, Moment, TimeUnit
from .offline import AsyncOfflineClock, OfflineClock
from .threaded import Clock

__all__ = [
    "AsyncClock",
    "ClockContext",
    "Moment",
    "OfflineClock",
    "AsyncOfflineClock",
    "Clock",
    "TimeUnit",
]



import collections

from supriya import CalculationRate
from supriya.synthdefs import UGen


class Gendy1(UGen):
    """
    A dynamic stochastic synthesis generator.

    ::

        >>> gendy_1 = supriya.ugens.Gendy1.ar(
        ...     adparam=1,
        ...     ampdist=1,
        ...     ampscale=0.5,
        ...     ddparam=1,
        ...     durdist=1,
        ...     durscale=0.5,
        ...     init_cps=12,
        ...     knum=10,
        ...     maxfrequency=660,
        ...     minfrequency=440,
        ... )
        >>> gendy_1
        Gendy1.ar()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [
            ("ampdist", 1),
            ("durdist", 1),
            ("adparam", 1),
            ("ddparam", 1),
            ("minfrequency", 440),
            ("maxfrequency", 660),
            ("ampscale", 0.5),
            ("durscale", 0.5),
            ("init_cps", 12),
            ("knum", None),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### INITIALIZER ###

    def __init__(
        self,
        calculation_rate=None,
        adparam=1,
        ampdist=1,
        ampscale=0.5,
        ddparam=1,
        durdist=1,
        durscale=0.5,
        init_cps=12,
        knum=None,
        maxfrequency=660,
        minfrequency=440,
    ):
        if knum is None:
            knum = init_cps
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            adparam=adparam,
            ampdist=ampdist,
            ampscale=ampscale,
            ddparam=ddparam,
            durdist=durdist,
            durscale=durscale,
            init_cps=init_cps,
            knum=knum,
            maxfrequency=maxfrequency,
            minfrequency=minfrequency,
        )


class Gendy2(UGen):
    """
    A dynamic stochastic synthesis generator.

    ::

        >>> gendy_2 = supriya.ugens.Gendy2.ar(
        ...     a=1.17,
        ...     adparam=1,
        ...     ampdist=1,
        ...     ampscale=0.5,
        ...     c=0.31,
        ...     ddparam=1,
        ...     durdist=1,
        ...     durscale=0.5,
        ...     init_cps=12,
        ...     knum=10,
        ...     maxfrequency=660,
        ...     minfrequency=440,
        ... )
        >>> gendy_2
        Gendy2.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("ampdist", 1),
            ("durdist", 1),
            ("adparam", 1),
            ("ddparam", 1),
            ("minfrequency", 440),
            ("maxfrequency", 660),
            ("ampscale", 0.5),
            ("durscale", 0.5),
            ("init_cps", 12),
            ("knum", None),
            ("a", 1.17),
            ("c", 0.31),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Gendy3(UGen):
    """
    A dynamic stochastic synthesis generator.

    ::

        >>> gendy_3 = supriya.ugens.Gendy3.ar(
        ...     adparam=1,
        ...     ampdist=1,
        ...     ampscale=0.5,
        ...     ddparam=1,
        ...     durdist=1,
        ...     durscale=0.5,
        ...     frequency=440,
        ...     init_cps=12,
        ...     knum=10,
        ... )
        >>> gendy_3
        Gendy3.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [
            ("ampdist", 1),
            ("durdist", 1),
            ("adparam", 1),
            ("ddparam", 1),
            ("frequency", 440),
            ("ampscale", 0.5),
            ("durscale", 0.5),
            ("init_cps", 12),
            ("knum", None),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



import abc

from supriya.system import SupriyaObject


class ServerObject(SupriyaObject):
    """
    A proxy of an object on a server.

    Server objects can be allocated and freed.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_server",)

    ### INITIALIZER ###

    @abc.abstractmethod
    def __init__(self):
        self._server = None

    ### PUBLIC METHODS ###

    @abc.abstractmethod
    def allocate(self, server):
        import supriya.realtime

        if self.is_allocated and self.server is server:
            return
        assert self.server is None, (self, self.server)
        assert isinstance(server, supriya.realtime.Server)
        assert server.is_running
        self._server = server

    @abc.abstractmethod
    def free(self):
        self._server = None

    ### PUBLIC PROPERTIES ###

    @abc.abstractproperty
    def is_allocated(self):
        raise NotImplementedError

    @property
    def server(self):
        return self._server



import uqbar.graphs

from supriya import BinaryOperator, UnaryOperator


class SynthDefGrapher:
    r"""
    Graphs SynthDefs.

    ..  container:: example

        ::

            >>> ugen_graph = supriya.ugens.LFNoise2.ar()
            >>> result = ugen_graph.transpose([0, 3, 7])

        ::

            >>> supriya.graph(result)  # doctest: +SKIP

        ::

            >>> print(format(result.__graph__(), "graphviz"))
            digraph synthdef_c481c3d42e3cfcee0267250247dab51f {
                graph [bgcolor=transparent,
                    color=lightslategrey,
                    dpi=72,
                    fontname=Arial,
                    outputorder=edgesfirst,
                    overlap=prism,
                    penwidth=2,
                    rankdir=LR,
                    ranksep=1,
                    splines=spline,
                    style="dotted, rounded"];
                node [fontname=Arial,
                    fontsize=12,
                    penwidth=2,
                    shape=Mrecord,
                    style="filled, rounded"];
                edge [penwidth=2];
                ugen_0 [fillcolor=lightsteelblue2,
                    label="<f_0> LFNoise2\n(audio) | { { <f_1_0_0> frequency:\n500.0 } | { <f_1_1_0> 0 } }"];
                ugen_1 [fillcolor=lightsteelblue2,
                    label="<f_0> UnaryOpUGen\n[HZ_TO_MIDI]\n(audio) | { { <f_1_0_0> source } | { <f_1_1_0> 0 } }"];
                ugen_2 [fillcolor=lightsteelblue2,
                    label="<f_0> UnaryOpUGen\n[MIDI_TO_HZ]\n(audio) | { { <f_1_0_0> source } | { <f_1_1_0> 0 } }"];
                ugen_3 [fillcolor=lightsteelblue2,
                    label="<f_0> BinaryOpUGen\n[ADDITION]\n(audio) | { { <f_1_0_0> left | <f_1_0_1> right:\n3.0 } | { <f_1_1_0> 0 } }"];
                ugen_4 [fillcolor=lightsteelblue2,
                    label="<f_0> UnaryOpUGen\n[MIDI_TO_HZ]\n(audio) | { { <f_1_0_0> source } | { <f_1_1_0> 0 } }"];
                ugen_5 [fillcolor=lightsteelblue2,
                    label="<f_0> BinaryOpUGen\n[ADDITION]\n(audio) | { { <f_1_0_0> left | <f_1_0_1> right:\n7.0 } | { <f_1_1_0> 0 } }"];
                ugen_6 [fillcolor=lightsteelblue2,
                    label="<f_0> UnaryOpUGen\n[MIDI_TO_HZ]\n(audio) | { { <f_1_0_0> source } | { <f_1_1_0> 0 } }"];
                ugen_0:f_1_1_0:e -> ugen_1:f_1_0_0:w [color=steelblue];
                ugen_1:f_1_1_0:e -> ugen_2:f_1_0_0:w [color=steelblue];
                ugen_1:f_1_1_0:e -> ugen_3:f_1_0_0:w [color=steelblue];
                ugen_1:f_1_1_0:e -> ugen_5:f_1_0_0:w [color=steelblue];
                ugen_3:f_1_1_0:e -> ugen_4:f_1_0_0:w [color=steelblue];
                ugen_5:f_1_1_0:e -> ugen_6:f_1_0_0:w [color=steelblue];
            }

    """

    ### PRIVATE METHODS ###

    @staticmethod
    def _connect_nodes(synthdef, ugen_node_mapping):
        import supriya.synthdefs

        for ugen in synthdef.ugens:
            tail_node = ugen_node_mapping[ugen]
            for i, input_ in enumerate(ugen.inputs):
                if not isinstance(input_, supriya.synthdefs.OutputProxy):
                    continue
                tail_field = tail_node["inputs"][i]
                source = input_.source
                head_node = ugen_node_mapping[source]
                head_field = head_node["outputs"][input_.output_index]
                edge = uqbar.graphs.Edge(head_port_position="w", tail_port_position="e")
                edge.attach(head_field, tail_field)
                if source.calculation_rate == supriya.CalculationRate.CONTROL:
                    edge.attributes["color"] = "goldenrod"
                elif source.calculation_rate == supriya.CalculationRate.AUDIO:
                    edge.attributes["color"] = "steelblue"
                else:
                    edge.attributes["color"] = "salmon"

    @staticmethod
    def _create_ugen_input_group(ugen, ugen_index):
        if not ugen.inputs:
            return None
        input_group = uqbar.graphs.RecordGroup(name="inputs")
        for i, input_ in enumerate(ugen.inputs):
            label = ""
            input_name = None
            if i < len(ugen._ordered_input_names):
                input_name = tuple(ugen._ordered_input_names)[i]
            if input_name:
                # input_name = r'\n'.join(input_name.split('_'))
                if isinstance(input_, float):
                    label = r"{}:\n{}".format(input_name, input_)
                else:
                    label = input_name
            elif isinstance(input_, float):
                label = str(input_)
            label = label or None
            field = uqbar.graphs.RecordField(
                label=label, name="ugen_{}_input_{}".format(ugen_index, i)
            )
            input_group.append(field)
        return input_group

    @staticmethod
    def _create_ugen_node_mapping(synthdef):
        import supriya.synthdefs

        ugen_node_mapping = {}
        for ugen in synthdef.ugens:
            ugen_index = synthdef.ugens.index(ugen)
            node = uqbar.graphs.Node(name="ugen_{}".format(ugen_index))
            if ugen.calculation_rate == supriya.CalculationRate.CONTROL:
                node.attributes["fillcolor"] = "lightgoldenrod2"
            elif ugen.calculation_rate == supriya.CalculationRate.AUDIO:
                node.attributes["fillcolor"] = "lightsteelblue2"
            else:
                node.attributes["fillcolor"] = "lightsalmon2"
            title_field = SynthDefGrapher._create_ugen_title_field(ugen)
            node.append(title_field)
            group = uqbar.graphs.RecordGroup()
            input_group = SynthDefGrapher._create_ugen_input_group(ugen, ugen_index)
            if input_group is not None:
                group.append(input_group)
            output_group = SynthDefGrapher._create_ugen_output_group(
                synthdef, ugen, ugen_index
            )
            if output_group is not None:
                group.append(output_group)
            node.append(group)
            ugen_node_mapping[ugen] = node
        return ugen_node_mapping

    @staticmethod
    def _create_ugen_output_group(synthdef, ugen, ugen_index):
        import supriya.ugens

        if not ugen.outputs:
            return None
        output_group = uqbar.graphs.RecordGroup(name="outputs")
        for i, output in enumerate(ugen.outputs):
            label = str(i)
            if isinstance(ugen, supriya.synthdefs.Control):
                parameter_index = ugen.special_index + i
                parameter = dict(synthdef.indexed_parameters)[parameter_index]
                parameter_name = parameter.name
                # parameter_name = r'\n'.join(parameter.name.split('_'))
                label = r"{}:\n{}".format(parameter_name, parameter.value)
            field = uqbar.graphs.RecordField(
                label=label, name="ugen_{}_output_{}".format(ugen_index, i)
            )
            output_group.append(field)
        return output_group

    @staticmethod
    def _create_ugen_title_field(ugen):
        import supriya.synthdefs
        import supriya.ugens

        name = type(ugen).__name__
        calculation_rate = ugen.calculation_rate.name.lower()
        label_template = r"{name}\n({calculation_rate})"
        operator = None
        if isinstance(ugen, supriya.synthdefs.BinaryOpUGen):
            operator = BinaryOperator(ugen.special_index).name
            label_template = r"{name}\n[{operator}]\n({calculation_rate})"
        elif isinstance(ugen, supriya.synthdefs.UnaryOpUGen):
            operator = UnaryOperator(ugen.special_index).name
            label_template = r"{name}\n[{operator}]\n({calculation_rate})"
        title_field = uqbar.graphs.RecordField(
            label=label_template.format(
                name=name, operator=operator, calculation_rate=calculation_rate
            )
        )
        return title_field

    @staticmethod
    def _style_graph(graph):
        graph.attributes.update(
            {
                "bgcolor": "transparent",
                "color": "lightslategrey",
                "dpi": 72,
                "fontname": "Arial",
                "outputorder": "edgesfirst",
                "overlap": "prism",
                "penwidth": 2,
                "rankdir": "LR",
                "ranksep": 1,
                "splines": "spline",
                "style": ("dotted", "rounded"),
            }
        )
        graph.edge_attributes.update({"penwidth": 2})
        graph.node_attributes.update(
            {
                "fontname": "Arial",
                "fontsize": 12,
                "penwidth": 2,
                "shape": "Mrecord",
                "style": ("filled", "rounded"),
            }
        )

    ### PUBLIC METHODS ###

    @staticmethod
    def graph(synthdef):
        import supriya.synthdefs

        assert isinstance(synthdef, supriya.synthdefs.SynthDef)
        graph = uqbar.graphs.Graph(name="synthdef_{}".format(synthdef.actual_name))
        ugen_node_mapping = SynthDefGrapher._create_ugen_node_mapping(synthdef)
        for node in sorted(ugen_node_mapping.values(), key=lambda x: x.name):
            graph.append(node)
        SynthDefGrapher._connect_nodes(synthdef, ugen_node_mapping)
        SynthDefGrapher._style_graph(graph)
        return graph



import asyncio
import logging
import re
import threading
from os import PathLike
from typing import Optional, Set, Union

from uqbar.objects import new

import supriya.exceptions
from supriya.commands import (  # type: ignore
    FailResponse,
    GroupNewRequest,
    GroupQueryTreeRequest,
    NotifyRequest,
    QuitRequest,
    SyncRequest,
)
from supriya.enums import CalculationRate, NodeAction
from supriya.exceptions import ServerOffline
from supriya.osc.protocols import (
    AsyncOscProtocol,
    HealthCheck,
    OscProtocolOffline,
    ThreadedOscProtocol,
)
from supriya.querytree import QueryTreeGroup, QueryTreeSynth
from supriya.scsynth import Options, find

from ..typing import AddActionLike, CalculationRateLike
from .allocators import BlockAllocator, NodeIdAllocator
from .buffers import Buffer, BufferGroup
from .buses import AudioInputBusGroup, AudioOutputBusGroup, Bus, BusGroup
from .meters import Meters
from .nodes import Group, Node, RootNode, Synth
from .protocols import AsyncProcessProtocol, SyncProcessProtocol
from .recorder import Recorder

try:
    from .shm import ServerSHM
except (ImportError, ModuleNotFoundError):
    ServerSHM = None

logger = logging.getLogger("supriya.server")

DEFAULT_IP_ADDRESS = "127.0.0.1"
DEFAULT_PORT = 57110


class BaseServer:

    ### INITIALIZER ###

    def __init__(self):
        # address
        self._ip_address = DEFAULT_IP_ADDRESS
        self._port = DEFAULT_PORT
        # process
        self._client_id = 0
        self._is_owner = False
        self._is_running = False
        self._latency = 0.1
        self._maximum_logins = 1
        self._options = Options()
        self._osc_protocol = None
        self._process_protocol = None
        self._status = None
        self._shm = None
        # allocators
        self._audio_bus_allocator = None
        self._buffer_allocator = None
        self._control_bus_allocator = None
        self._node_id_allocator = None
        self._sync_id = 0
        # proxy mappings
        self._synthdefs = {}

    ### SPECIAL METHODS ###

    def __repr__(self):
        if not self.is_running:
            return f"<{type(self).__name__}: offline>"
        string = "<{name}: {protocol}://{ip}:{port}, "
        string += "{inputs}i{outputs}o>"
        return string.format(
            name=type(self).__name__,
            protocol=self.options.protocol,
            ip=self.ip_address,
            port=self.port,
            inputs=self.options.input_bus_channel_count,
            outputs=self.options.output_bus_channel_count,
        )

    ### PRIVATE METHODS ###

    def _handle_failed_response(self, message):
        logger.warning("Fail: {}".format(message))

    def _handle_status_reply_response(self, message):
        from supriya.commands import Response

        response = Response.from_osc_message(message)
        self._status = response

    def _handle_synthdef_removed_response(self, message):
        from supriya.commands import Response

        response = Response.from_osc_message(message)
        synthdef_name = response.synthdef_name
        self._synthdefs.pop(synthdef_name, None)

    def _setup_allocators(self):
        self._audio_bus_allocator = BlockAllocator(
            heap_maximum=self._options.audio_bus_channel_count,
            heap_minimum=self._options.first_private_bus_id,
        )
        self._buffer_allocator = BlockAllocator(heap_maximum=self._options.buffer_count)
        self._control_bus_allocator = BlockAllocator(
            heap_maximum=self._options.control_bus_channel_count
        )
        self._node_id_allocator = NodeIdAllocator(
            initial_node_id=self._options.initial_node_id, client_id=self.client_id
        )
        self._sync_id = self.client_id << 26

    def _setup_osc_callbacks(self):
        self._osc_protocol.register(
            pattern="/d_removed", procedure=self._handle_synthdef_removed_response
        )
        self._osc_protocol.register(
            pattern="/status.reply", procedure=self._handle_status_reply_response
        )
        self._osc_protocol.register(
            pattern="/fail", procedure=self._handle_failed_response
        )

    def _setup_shm(self):
        if ServerSHM is None:
            return
        self._shm = ServerSHM(self.port, self.options.control_bus_channel_count)

    def _teardown_allocators(self):
        self._audio_bus_allocator = None
        self._buffer_allocator = None
        self._control_bus_allocator = None
        self._node_id_allocator = None
        self._sync_id = 0

    def _teardown_shm(self):
        self._shm = None

    ### PUBLIC METHODS ###

    def send(self, message):
        if not message:
            raise ValueError
        if not self.is_running:
            raise ServerOffline
        self._osc_protocol.send(message)
        return self

    ### PUBLIC PROPERTIES ###

    @property
    def audio_bus_allocator(self):
        return self._audio_bus_allocator

    @property
    def buffer_allocator(self):
        return self._buffer_allocator

    @property
    def client_id(self) -> int:
        return self._client_id

    @property
    def control_bus_allocator(self):
        return self._control_bus_allocator

    @property
    def ip_address(self) -> str:
        return self._ip_address

    @property
    def is_owner(self) -> bool:
        return self._is_owner

    @property
    def is_running(self) -> bool:
        return self._is_running

    @property
    def latency(self) -> float:
        return self._latency

    @latency.setter
    def latency(self, latency):
        self._latency = float(latency)

    @property
    def maximum_logins(self) -> int:
        return self._maximum_logins

    @property
    def next_sync_id(self) -> int:
        sync_id = self._sync_id
        self._sync_id += 1
        return sync_id

    @property
    def node_id_allocator(self):
        return self._node_id_allocator

    @property
    def osc_protocol(self):
        return self._osc_protocol

    @property
    def options(self) -> Options:
        return self._options

    @property
    def port(self) -> int:
        return self._port

    @property
    def process_protocol(self):
        return self._process_protocol

    @property
    def status(self):
        return self._status


class AsyncServer(BaseServer):

    ### CLASS VARIABLES ###

    _servers: Set["AsyncServer"] = set()

    ### INTIALIZER ###

    def __init__(self):
        BaseServer.__init__(self)
        self._boot_future = None
        self._quit_future = None

    ### SPECIAL METHODS ###

    def __contains__(self, expr):
        if isinstance(expr, supriya.synthdefs.SynthDef):
            name = expr.actual_name
            if name in self._synthdefs and self._synthdefs[name] == expr:
                return True
        return False

    ### PRIVATE METHODS ###

    async def _connect(self):
        self._osc_protocol = AsyncOscProtocol()
        await self._osc_protocol.connect(
            ip_address=self._ip_address,
            port=self._port,
            healthcheck=HealthCheck(
                request_pattern=["/status"],
                response_pattern=["/status.reply"],
                callback=self._shutdown,
                max_attempts=5,
                timeout=1.0,
                backoff_factor=1.5,
            ),
        )
        self._is_running = True
        self._setup_osc_callbacks()
        await self._setup_notifications()
        self._setup_allocators()
        if self.client_id == 0:
            await self._setup_default_groups()
            await self._setup_system_synthdefs()
        self.boot_future.set_result(True)
        self._servers.add(self)

    async def _disconnect(self):
        self._is_running = False
        self._is_owner = False
        self._client_id = None
        self._maximum_logins = None
        self._teardown_shm()
        await self._osc_protocol.disconnect()
        await self._osc_protocol.exit_future
        self._teardown_allocators()
        if self in self._servers:
            self._servers.remove(self)
        self.quit_future.set_result(True)
        if not self.boot_future.done():
            self.boot_future.set_result(False)

    async def _setup_default_groups(self):
        request = GroupNewRequest(
            items=[
                GroupNewRequest.Item(1, i, 0) for i in range(1, self.maximum_logins + 1)
            ]
        )
        self.send(request.to_osc())

    async def _setup_notifications(self):
        request = NotifyRequest(True)
        response = await request.communicate_async(server=self)
        if isinstance(response, FailResponse):
            await self._shutdown()
            raise supriya.exceptions.TooManyClients
        self._client_id, self._maximum_logins = response.action[1], response.action[2]

    async def _setup_system_synthdefs(self):
        pass

    async def _shutdown(self):
        if not self.is_running:
            return
        elif self.is_owner:
            await self.quit()
        else:
            await self.disconnect()

    ### PUBLIC METHODS ###

    async def boot(
        self,
        *,
        ip_address: str = DEFAULT_IP_ADDRESS,
        port: int = DEFAULT_PORT,
        scsynth_path: Optional[str] = None,
        options: Optional[Options] = None,
        **kwargs,
    ) -> "AsyncServer":
        if self._is_running:
            raise supriya.exceptions.ServerOnline
        port = port or DEFAULT_PORT
        loop = asyncio.get_running_loop()
        self._boot_future = loop.create_future()
        self._quit_future = loop.create_future()
        self._options = new(options or Options(), **kwargs)
        scsynth_path = find(scsynth_path)
        self._process_protocol = AsyncProcessProtocol()
        await self._process_protocol.boot(self._options, scsynth_path, port)
        if not await self._process_protocol.boot_future:
            self._boot_future.set_result(False)
            self._quit_future.set_result(True)
            raise supriya.exceptions.ServerCannotBoot
        self._ip_address = ip_address
        self._is_owner = True
        self._port = port
        await self._connect()
        return self

    async def connect(
        self, *, ip_address: str = DEFAULT_IP_ADDRESS, port: int = DEFAULT_PORT
    ) -> "AsyncServer":
        if self._is_running:
            raise supriya.exceptions.ServerOnline
        loop = asyncio.get_running_loop()
        self._boot_future = loop.create_future()
        self._quit_future = loop.create_future()
        self._ip_address = ip_address
        self._is_owner = False
        self._port = port
        await self._connect()
        return self

    async def disconnect(self) -> "AsyncServer":
        if not self._is_running:
            raise ServerOffline
        if self._is_owner:
            raise supriya.exceptions.OwnedServerShutdown(
                "Cannot disconnect from owned server."
            )
        await self._disconnect()
        return self

    async def query(self, include_controls=True) -> QueryTreeGroup:
        request = GroupQueryTreeRequest(node_id=0, include_controls=include_controls)
        response = await request.communicate_async(server=self)
        return response.query_tree_group

    async def quit(self, force: bool = False) -> "AsyncServer":
        if not self._is_running:
            return self
        if not self._is_owner and not force:
            raise supriya.exceptions.UnownedServerShutdown(
                "Cannot quit unowned server without force flag."
            )
        try:
            await QuitRequest().communicate_async(server=self, sync=True, timeout=1)
        except (OscProtocolOffline, asyncio.TimeoutError):
            pass
        if self._process_protocol is not None:
            self._process_protocol.quit()
        await self._disconnect()
        return self

    ### PUBLIC PROPERTIES ###

    @property
    def boot_future(self):
        return self._boot_future

    @property
    def default_group(self) -> int:
        return self.client_id + 1

    @property
    def quit_future(self):
        return self._quit_future


class Server(BaseServer):
    """
    An scsynth server proxy.

    ::

        >>> import supriya.realtime
        >>> server = supriya.realtime.Server()
        >>> server.boot()
        <Server: udp://127.0.0.1:57110, 8i8o>

    ::

        >>> server.quit()
        <Server: offline>

    """

    ### CLASS VARIABLES ###

    _servers: Set["Server"] = set()

    ### INITIALIZER ###

    def __init__(self):
        BaseServer.__init__(self)
        self._lock = threading.RLock()
        # proxies
        self._audio_input_bus_group = None
        self._audio_output_bus_group = None
        self._default_group = None
        self._root_node = None
        self._meters = Meters(self)
        self._recorder = Recorder(self)
        # proxy mappings
        self._audio_buses = {}
        self._buffer_proxies = {}
        self._buffers = {}
        self._control_bus_proxies = {}
        self._control_buses = {}
        self._nodes = {}
        self._pending_synths = {}

    ### SPECIAL METHODS ###

    def __contains__(self, expr):
        import supriya.realtime
        import supriya.synthdefs

        if isinstance(expr, supriya.realtime.Node):
            if expr.server is not self:
                return False
            node_id = expr.node_id
            if node_id in self._nodes and self._nodes[node_id] is expr:
                return True
        elif isinstance(expr, supriya.synthdefs.SynthDef):
            name = expr.actual_name
            if name in self._synthdefs and self._synthdefs[name] == expr:
                return True
        elif isinstance(expr, supriya.realtime.ServerObject):
            return expr.server is self
        return False

    def __getitem__(self, item: Union[int, str]) -> Union[Buffer, Bus, Node]:
        """
        Get ``item`` from server.

        ::

            >>> server = supriya.Server().boot()
            >>> supriya.Synth(name="foo").allocate(server)
            <+ Synth: 1000 default (foo)>

        ::

            >>> server[1000]
            <+ Synth: 1000 default (foo)>

        ::

            >>> server["foo"]
            <+ Synth: 1000 default (foo)>

        ::

            >>> server["b10"]
            <+ Buffer: 10, 1ch, 1>

        ::

            >>> server["a0"]
            <+ Bus: 0 (audio)>

        ::

            >>> server["c16"]
            <+ Bus: 16 (control)>

        ::

            >>> server = server.quit()
            >>> server["c16"]
            Traceback (most recent call last):
            ...
            supriya.exceptions.ServerOffline

        """
        import supriya

        if not self.is_running:
            raise ServerOffline
        if isinstance(item, str):
            match = re.match(r"b(?P<id>\d+)", item)
            if match:
                id_ = int(match.groupdict()["id"])
                return supriya.realtime.Buffer(id_).allocate(server=self)
            match = re.match(r"c(?P<id>\d+)", item)
            if match:
                id_ = int(match.groupdict()["id"])
                return supriya.realtime.Bus(id_, "control").allocate(server=self)
            match = re.match(r"a(?P<id>\d+)", item)
            if match:
                id_ = int(match.groupdict()["id"])
                return supriya.realtime.Bus(id_, "audio").allocate(server=self)
            if self.root_node is None:
                raise ServerOffline
            result = self.root_node[item]
        elif isinstance(item, int):
            result = self._nodes.get(item)
        else:
            raise ValueError(item)
        if isinstance(result, set) and len(result) == 1:
            return tuple(result)[0]
        return result

    def __graph__(self):
        """
        Graph server.

        ::

            >>> server = supriya.Server().boot()
            >>> group = supriya.Group(
            ...     [
            ...         supriya.Synth(),
            ...         supriya.Group(
            ...             [
            ...                 supriya.Synth(),
            ...                 supriya.Synth(),
            ...             ]
            ...         ),
            ...     ]
            ... ).allocate(server)

        ::

            >>> graph = server.__graph__()
            >>> print(format(graph, "graphviz"))
            digraph G {
                graph [bgcolor=transparent,
                    color=lightslategrey,
                    dpi=72,
                    fontname=Arial,
                    outputorder=edgesfirst,
                    overlap=prism,
                    penwidth=2,
                    rankdir=TB,
                    ranksep=0.5,
                    splines=spline,
                    style="dotted, rounded"];
                node [fontname=Arial,
                    fontsize=12,
                    penwidth=2,
                    shape=Mrecord,
                    style="filled, rounded"];
                edge [penwidth=2];
                "root-node-0" [fillcolor=lightsalmon2,
                    label="{ <f_0_0> RootNode | <f_0_1> id: 0 }"];
                "group-1" [fillcolor=lightsteelblue2,
                    label="{ <f_0_0> Group | <f_0_1> id: 1 }"];
                "group-1000" [fillcolor=lightsteelblue2,
                    label="{ <f_0_0> Group | <f_0_1> id: 1000 }"];
                "synth-1001" [fillcolor=lightgoldenrod2,
                    label="{ <f_0_0> Synth | <f_0_1> id: 1001 }"];
                "group-1002" [fillcolor=lightsteelblue2,
                    label="{ <f_0_0> Group | <f_0_1> id: 1002 }"];
                "synth-1003" [fillcolor=lightgoldenrod2,
                    label="{ <f_0_0> Synth | <f_0_1> id: 1003 }"];
                "synth-1004" [fillcolor=lightgoldenrod2,
                    label="{ <f_0_0> Synth | <f_0_1> id: 1004 }"];
                "root-node-0" -> "group-1";
                "group-1" -> "group-1000";
                "group-1000" -> "synth-1001";
                "group-1000" -> "group-1002";
                "group-1002" -> "synth-1003";
                "group-1002" -> "synth-1004";
            }

        ::

            >>> supriya.graph(server)  # doctest: +SKIP

        """

        return self.root_node.__graph__()

    ### PRIVATE METHODS ###

    def _connect(self):
        self._osc_protocol = ThreadedOscProtocol()
        self._osc_protocol.connect(
            ip_address=self.ip_address,
            port=self.port,
            healthcheck=HealthCheck(
                request_pattern=["/status"],
                response_pattern=["/status.reply"],
                callback=self._shutdown,
                max_attempts=5,
                timeout=1.0,
                backoff_factor=1.5,
            ),
        )
        self._is_running = True
        self._setup_osc_callbacks()
        self._setup_notifications()
        self._setup_allocators()
        self._setup_proxies()
        if self.client_id == 0:
            self._setup_default_groups()
            self._setup_system_synthdefs()
        self._servers.add(self)

    def _disconnect(self):
        logger.info("disconnecting")
        self._is_running = False
        self._is_owner = False
        self._client_id = None
        self._maximum_logins = None
        self._teardown_shm()
        self._osc_protocol.disconnect()
        self._teardown_proxies()
        self._teardown_allocators()
        if self in self._servers:
            self._servers.remove(self)
        logger.info("disconnected")

    def _get_buffer_proxy(self, buffer_id):
        import supriya.realtime

        buffer_proxy = self._buffer_proxies.get(buffer_id)
        if not buffer_proxy:
            buffer_proxy = supriya.realtime.BufferProxy(
                buffer_id=buffer_id, server=self
            )
            self._buffer_proxies[buffer_id] = buffer_proxy
        return buffer_proxy

    def _get_control_bus_proxy(self, bus_id):
        import supriya.realtime
        import supriya.synthdefs

        control_bus_proxy = self._control_bus_proxies.get(bus_id)
        if not control_bus_proxy:
            control_bus_proxy = supriya.realtime.BusProxy(
                bus_id=bus_id,
                calculation_rate=supriya.CalculationRate.CONTROL,
                server=self,
            )
            self._control_bus_proxies[bus_id] = control_bus_proxy
        return control_bus_proxy

    def _handle_buffer_info_response(self, message):
        from supriya.commands import Response

        response = Response.from_osc_message(message)
        for item in response.items:
            buffer_proxy = self._get_buffer_proxy(item.buffer_id)
            if buffer_proxy:
                buffer_proxy._handle_response(item)

    def _handle_control_bus_set_response(self, message):
        from supriya.commands import Response

        response = Response.from_osc_message(message)
        for item in response:
            bus_id = item.bus_id
            bus_proxy = self._get_control_bus_proxy(bus_id)
            bus_proxy._value = item.bus_value

    def _handle_control_bus_setn_response(self, message):
        from supriya.commands import Response

        response = Response.from_osc_message(message)
        for item in response:
            starting_bus_id = item.starting_bus_id
            for i, value in enumerate(item.bus_values):
                bus_id = starting_bus_id + i
                bus_proxy = self._get_control_bus_proxy(bus_id)
                bus_proxy._value = value

    def _handle_node_info_response(self, message):
        from supriya.commands import Response
        from supriya.realtime import Group, Synth

        response = Response.from_osc_message(message)
        with self._lock:
            node_id = response.node_id
            node = self._nodes.get(node_id)
            if node is not None:
                node._handle_response(response)
            elif response.action == NodeAction.NODE_CREATED:
                if response.is_group:
                    node = Group()
                else:
                    node = self._pending_synths.pop(node_id, Synth())
                node._register_with_local_server(self, node_id=response.node_id)
                parent = self._nodes[response.parent_id]
                node._set_parent(parent)
                if response.previous_node_id:
                    previous_child = self._nodes[response.previous_node_id]
                    index = parent.index(previous_child)
                    parent._children.insert(index + 1, node)
                else:
                    parent._children.append(node)

    def _handle_synthdef_removed_response(self, message):
        from supriya.commands import Response

        response = Response.from_osc_message(message)
        synthdef_name = response.synthdef_name
        self._synthdefs.pop(synthdef_name, None)

    def _rehydrate(self):
        from supriya.realtime import Group, Synth

        def recurse(query_tree_node, node):
            for query_tree_child in query_tree_node.children:
                if isinstance(query_tree_child, QueryTreeGroup):
                    group = Group()
                    group._register_with_local_server(
                        self, node_id=query_tree_child.node_id
                    )
                    node._children.append(group)
                    recurse(query_tree_child, group)
                elif isinstance(query_tree_child, QueryTreeSynth):
                    synth = Synth()
                    synth._register_with_local_server(
                        self, node_id=query_tree_child.node_id
                    )
                    node._children.append(synth)
                    for query_tree_control in query_tree_child.children:
                        pass

        recurse(self.query(), self.root_node)

    def _setup_notifications(self):
        request = NotifyRequest(True)
        response = request.communicate(server=self)
        if isinstance(response, FailResponse):
            self._shutdown()
            raise supriya.exceptions.TooManyClients
        if len(response.action) == 2:  # supernova doesn't provide a max logins value
            self._client_id, self._maximum_logins = response.action[1], 1
        else:
            self._client_id, self._maximum_logins = response.action[1:3]

    def _setup_default_groups(self):
        default_groups = [
            supriya.Group(node_id_is_permanent=True) for _ in range(self.maximum_logins)
        ]
        self.root_node.extend(default_groups)
        self._default_group = default_groups[self.client_id]

    def _setup_proxies(self):
        self._audio_input_bus_group = AudioInputBusGroup(self)
        self._audio_output_bus_group = AudioOutputBusGroup(self)
        self._root_node = supriya.realtime.RootNode(server=self)
        self._nodes[0] = self._root_node

    def _setup_osc_callbacks(self):
        super()._setup_osc_callbacks()
        self._osc_protocol.register(
            pattern="/b_info", procedure=self._handle_buffer_info_response
        )
        self._osc_protocol.register(
            pattern="/c_set", procedure=self._handle_control_bus_set_response
        )
        self._osc_protocol.register(
            pattern="/c_setn", procedure=self._handle_control_bus_setn_response
        )
        for pattern in (
            "/n_end",
            "/n_go",
            "/n_info",
            "/n_move",
            "/n_off",
            "/n_on",
            "/n_set",
            "/n_setn",
        ):
            self._osc_protocol.register(
                pattern=pattern, procedure=self._handle_node_info_response
            )

    def _setup_system_synthdefs(self, local_only=False):
        import supriya.assets.synthdefs
        import supriya.synthdefs

        system_synthdefs = []
        for name in dir(supriya.assets.synthdefs):
            if not name.startswith("system_"):
                continue
            system_synthdef = getattr(supriya.assets.synthdefs, name)
            if not isinstance(system_synthdef, supriya.synthdefs.SynthDef):
                continue
            system_synthdefs.append(system_synthdef)
        if local_only:
            for synthdef in system_synthdefs:
                synthdef._register_with_local_server(self)
        else:
            supriya.synthdefs.SynthDef._allocate_synthdefs(system_synthdefs, self)

    def _teardown_proxies(self):
        for set_ in tuple(self._audio_buses.values()):
            for x in tuple(set_):
                x.free()
        for set_ in tuple(self._control_buses.values()):
            for x in tuple(set_):
                x.free()
        for set_ in tuple(self._buffers.values()):
            for x in tuple(set_):
                x.free()
        for x in tuple(self._nodes.values()):
            x.free()
        self._audio_buses.clear()
        self._audio_input_bus_group = None
        self._audio_output_bus_group = None
        self._buffers.clear()
        self._buffer_proxies.clear()
        self._control_buses.clear()
        self._control_bus_proxies.clear()
        self._default_group = None
        self._nodes.clear()
        self._root_node = None
        self._synthdefs.clear()

    def _shutdown(self):
        if not self.is_running:
            return
        logger.info("shutting down")
        if self.is_owner:
            self.quit()
        else:
            self.disconnect()

    ### PUBLIC METHODS ###

    def add_buffer(
        self,
        channel_count: int = None,
        frame_count: int = None,
        starting_frame: int = None,
        file_path: Optional[PathLike] = None,
    ) -> Buffer:
        """
        Add a buffer.

        ::

            >>> server = supriya.Server().boot()
            >>> server.add_buffer(channel_count=2, frame_count=1024)
            <+ Buffer: 0, 2ch, 1024>

        """
        buffer_ = Buffer()
        if file_path:
            channel_indices = None
            if channel_count:
                channel_indices = tuple(range(channel_count))
            buffer_.allocate_from_file(
                self,
                file_path,
                channel_indices=channel_indices,
                frame_count=frame_count,
                starting_frame=starting_frame,
            )
        else:
            buffer_.allocate(
                channel_count=channel_count, frame_count=frame_count, server=self
            )
        return buffer_

    def add_buffer_group(
        self, buffer_count: int = 1, channel_count: int = None, frame_count: int = None
    ) -> BufferGroup:
        """
        Add a buffer group.

        ::

            >>> server = supriya.Server().boot()
            >>> server.add_buffer_group(buffer_count=8, channel_count=1, frame_count=1024)
            <+ BufferGroup{8}: 0>

        """
        buffer_group = BufferGroup(buffer_count)
        buffer_group.allocate(
            channel_count=channel_count, frame_count=frame_count, server=self
        )
        return buffer_group

    def add_bus(
        self, calculation_rate: CalculationRateLike = CalculationRate.CONTROL
    ) -> Bus:
        """
        Add a bus.

        ::

            >>> server = supriya.Server().boot()
            >>> server.add_bus()
            <+ Bus: 0 (control)>

        """
        bus = Bus(calculation_rate=calculation_rate)
        bus.allocate(server=self)
        return bus

    def add_bus_group(
        self,
        bus_count: int = 1,
        calculation_rate: CalculationRateLike = CalculationRate.CONTROL,
    ) -> BusGroup:
        """
        Add a bus group.

        ::

            >>> server = supriya.Server().boot()
            >>> server.add_bus_group(4, "audio")
            <+ BusGroup{4}: 16 (audio)>

        """
        bus_group = BusGroup(bus_count=bus_count, calculation_rate=calculation_rate)
        bus_group.allocate(server=self)
        return bus_group

    def add_group(self, add_action: AddActionLike = None) -> Group:
        """
        Add a group relative to the default group via ``add_action``.

        ::

            >>> server = supriya.Server().boot()
            >>> print(server.query())
            NODE TREE 0 group
                1 group

        ::

            >>> group = server.add_group()
            >>> print(server.query())
            NODE TREE 0 group
                1 group
                    1000 group

        """
        if self.default_group is None:
            raise ServerOffline
        return self.default_group.add_group(add_action=add_action)

    def add_synth(
        self, synthdef=None, add_action: AddActionLike = None, **kwargs
    ) -> Synth:
        """
        Add a synth relative to the default group via ``add_action``.

        ::

            >>> server = supriya.Server().boot()
            >>> print(server.query())
            NODE TREE 0 group
                1 group

        ::

            >>> synth = server.add_synth()
            >>> print(server.query())
            NODE TREE 0 group
                1 group
                    1000 default
                        out: 0.0, amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5

        """
        if self.default_group is None:
            raise ServerOffline
        return self.default_group.add_synth(
            synthdef=synthdef, add_action=add_action, **kwargs
        )

    def add_synthdef(self, synthdef) -> "Server":
        synthdef.allocate(server=self)
        return self

    def boot(
        self,
        *,
        ip_address: str = DEFAULT_IP_ADDRESS,
        port: int = DEFAULT_PORT,
        scsynth_path: Optional[str] = None,
        options: Optional[Options] = None,
        **kwargs,
    ) -> "Server":
        if self.is_running:
            raise supriya.exceptions.ServerOnline
        port = port or DEFAULT_PORT
        self._options = new(options or Options(), **kwargs)
        scsynth_path = find(scsynth_path)
        self._process_protocol = SyncProcessProtocol()
        self._process_protocol.boot(self._options, scsynth_path, port)
        self._ip_address = ip_address
        self._is_owner = True
        self._port = port
        self._connect()
        return self

    def connect(
        self, *, ip_address: str = DEFAULT_IP_ADDRESS, port: int = DEFAULT_PORT
    ) -> "Server":
        if self.is_running:
            raise supriya.exceptions.ServerOnline
        self._ip_address = ip_address
        self._is_owner = False
        self._port = port
        self._connect()
        if self.client_id > 0:
            self._setup_system_synthdefs(local_only=True)
            self._rehydrate()
        self._default_group = self._nodes[self.client_id + 1]
        return self

    def disconnect(self) -> "Server":
        if not self.is_running:
            raise ServerOffline
        if self._is_owner:
            raise supriya.exceptions.OwnedServerShutdown(
                "Cannot disconnect from owned server."
            )
        self._disconnect()
        return self

    def quit(self, force: bool = False) -> "Server":
        if not self.is_running:
            return self
        if not self._is_owner and not force:
            raise supriya.exceptions.UnownedServerShutdown(
                "Cannot quit unowned server without force flag."
            )
        if self.recorder.is_recording:
            self.recorder.stop()
        try:
            QuitRequest().communicate(server=self)
        except OscProtocolOffline:
            pass
        if self._process_protocol is not None:
            self._process_protocol.quit()
        self._disconnect()
        return self

    def query(self, include_controls=True) -> QueryTreeGroup:
        request = GroupQueryTreeRequest(node_id=0, include_controls=include_controls)
        response = request.communicate(server=self)
        return response.query_tree_group

    def reboot(self, options: Optional[Options] = None, **kwargs) -> "Server":
        self.quit()
        self.boot(options=options, **kwargs)
        return self

    def reset(self) -> "Server":
        self.send(["/d_freeAll"])
        self.send(["/g_freeAll", 0])
        self.send(["/clearSched"])
        self.sync()
        self._teardown_proxies()
        self._teardown_allocators()
        self._setup_allocators()
        self._setup_proxies()
        self._setup_default_groups()
        self._setup_system_synthdefs()
        self.sync()
        self._sync_id = 0
        return self

    def sync(self, sync_id: Optional[int] = None) -> "Server":
        if not self.is_running:
            return self
        if sync_id is None:
            sync_id = self.next_sync_id
        request = SyncRequest(sync_id=sync_id)
        request.communicate(server=self)
        return self

    ### PUBLIC PROPERTIES ###

    @property
    def audio_input_bus_group(self) -> Optional[AudioInputBusGroup]:
        return self._audio_input_bus_group

    @property
    def audio_output_bus_group(self) -> Optional[AudioOutputBusGroup]:
        return self._audio_output_bus_group

    @property
    def default_group(self) -> Optional[Group]:
        return self._default_group

    @property
    def meters(self):
        return self._meters

    @property
    def recorder(self):
        return self._recorder

    @property
    def root_node(self) -> Optional[RootNode]:
        return self._root_node



import supriya.synthdefs
import supriya.ugens


def _build_kick_synthdef():

    builder = supriya.synthdefs.SynthDefBuilder(out=0)

    with builder:

        ### ENVELOPE ###

        gate = supriya.ugens.Impulse.ar(frequency=2)
        envelope = supriya.synthdefs.Envelope.percussive(
            attack_time=0.01, release_time=1.0
        )
        envelope = supriya.ugens.EnvGen.ar(done_action=0, envelope=envelope, gate=gate)
        envelope = envelope.squared()

        ### NOISE COMPONENT ###

        noise = supriya.ugens.PinkNoise.ar()
        noise = supriya.ugens.BPF.ar(
            source=noise,
            frequency=supriya.ugens.LinLin.ar(
                source=envelope.cubed(), output_minimum=30, output_maximum=120
            ),
            reciprocal_of_q=2,
        )
        noise *= envelope

        ### PITCHED COMPONENT ###

        pitch = supriya.ugens.SinOsc.ar(
            frequency=supriya.ugens.LinLin.ar(
                source=envelope, output_minimum=10, output_maximum=80
            )
        )
        pitch = pitch * 2.0
        pitch = pitch.distort()
        pitch = supriya.ugens.RLPF.ar(
            source=pitch,
            frequency=supriya.ugens.LinLin.ar(
                source=envelope, output_minimum=30, output_maximum=120
            ),
            reciprocal_of_q=0.5,
        )
        pitch *= envelope

        mix = pitch + noise

        supriya.ugens.Out.ar(builder["out"], (mix, mix))

    synthdef = builder.build()
    return synthdef


kick = _build_kick_synthdef()

__all__ = ("kick",)



import collections
import math

from supriya import CalculationRate
from supriya.synthdefs import MultiOutUGen, PseudoUGen, UGen
from supriya.ugens.basic import Mix


class Balance2(MultiOutUGen):
    """
    A stereo signal balancer.

    ::

        >>> left = supriya.ugens.WhiteNoise.ar()
        >>> right = supriya.ugens.SinOsc.ar()
        >>> balance_2 = supriya.ugens.Balance2.ar(
        ...     left=left,
        ...     level=1,
        ...     position=0,
        ...     right=right,
        ... )
        >>> balance_2
        UGenArray({2})

    """

    _default_channel_count = 2
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict(
        [("left", None), ("right", None), ("position", 0.0), ("level", 1.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class BiPanB2(MultiOutUGen):
    """
    A 2D ambisonic b-format panner.

    ::

        >>> in_a = supriya.ugens.SinOsc.ar()
        >>> in_b = supriya.ugens.WhiteNoise.ar()
        >>> bi_pan_b_2 = supriya.ugens.BiPanB2.ar(
        ...     azimuth=-0.5,
        ...     gain=1,
        ...     in_a=in_a,
        ...     in_b=in_b,
        ... )
        >>> bi_pan_b_2
        UGenArray({3})

    ::

        >>> w, x, y = bi_pan_b_2

    """

    _default_channel_count = 3
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict(
        [("in_a", None), ("in_b", None), ("azimuth", None), ("gain", 1.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class DecodeB2(MultiOutUGen):
    """
    A 2D Ambisonic B-format decoder.

    ::

        >>> source = supriya.ugens.PinkNoise.ar()
        >>> w, x, y = supriya.ugens.PanB2.ar(
        ...     source=source,
        ...     azimuth=supriya.ugens.SinOsc.kr(),
        ... )
        >>> channel_count = 4
        >>> decode_b_2 = supriya.ugens.DecodeB2.ar(
        ...     channel_count=channel_count,
        ...     orientation=0.5,
        ...     w=w,
        ...     x=x,
        ...     y=y,
        ... )
        >>> decode_b_2
        UGenArray({4})

    """

    _default_channel_count = 4
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict(
        [("w", None), ("x", None), ("y", None), ("orientation", 0.5)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Pan2(MultiOutUGen):
    """
    A two channel equal power panner.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> pan_2 = supriya.ugens.Pan2.ar(
        ...     source=source,
        ... )
        >>> pan_2
        UGenArray({2})

    """

    _default_channel_count = 2
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("position", 0.0), ("level", 1.0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO,)


class Pan4(MultiOutUGen):
    """
    A four-channel equal-power panner.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pan_4 = supriya.ugens.Pan4.ar(
        ...     gain=1,
        ...     source=source,
        ...     x_position=0,
        ...     y_position=0,
        ... )
        >>> pan_4
        UGenArray({4})

    """

    _default_channel_count = 4
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("x_position", 0), ("y_position", 0), ("gain", 1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class PanAz(MultiOutUGen):
    """
    A multi-channel equal-power panner.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pan_az = supriya.ugens.PanAz.ar(
        ...     channel_count=8,
        ...     amplitude=1,
        ...     orientation=0.5,
        ...     position=0,
        ...     source=source,
        ...     width=2,
        ... )
        >>> pan_az
        UGenArray({8})

    """

    _default_channel_count = 1
    _has_settable_channel_count = True
    _ordered_input_names = collections.OrderedDict(
        [
            ("source", None),
            ("position", 0),
            ("amplitude", 1),
            ("width", 2),
            ("orientation", 0.5),
        ]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class PanB(MultiOutUGen):
    """
    A 3D ambisonic b-format panner.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pan_b = supriya.ugens.PanB.ar(
        ...     azimuth=0,
        ...     elevation=0,
        ...     gain=1,
        ...     source=source,
        ... )
        >>> pan_b
        UGenArray({3})

    """

    _default_channel_count = 3
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("azimuth", 0), ("elevation", 0), ("gain", 1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class PanB2(MultiOutUGen):
    """
    A 2D ambisonic b-format panner.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> pan_b_2 = supriya.ugens.PanB2.ar(
        ...     azimuth=0,
        ...     gain=1,
        ...     source=source,
        ... )
        >>> pan_b_2
        UGenArray({3})

    """

    _default_channel_count = 3
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict(
        [("source", None), ("azimuth", 0), ("gain", 1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Rotate2(MultiOutUGen):
    """
    Equal-power sound-field rotator.

    ::

        >>> x = supriya.ugens.PinkNoise.ar() * 0.4
        >>> y = supriya.ugens.LFTri.ar(frequency=880)
        >>> y *= supriya.ugens.LFPulse.kr(frequency=3, width=0.1)
        >>> position = supriya.ugens.LFSaw.kr(frequency=0.1)
        >>> rotate_2 = supriya.ugens.Rotate2.ar(
        ...     x=x,
        ...     y=y,
        ...     position=position,
        ... )
        >>> rotate_2
        UGenArray({2})

    Returns an array of the rotator's left and right outputs.
    """

    _default_channel_count = 2
    _has_settable_channel_count = False
    _ordered_input_names = collections.OrderedDict(
        [("x", None), ("y", None), ("position", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Splay(PseudoUGen):
    """
    A stereo signal spreader.

    ::

        >>> source = supriya.ugens.SinOsc.ar(frequency=[333, 444, 555, 666, 777])
        >>> splay = supriya.ugens.Splay.ar(source=source)
        >>> splay
        UGenArray({2})

    ::

        >>> print(splay)
        synthdef:
            name: ...
            ugens:
            -   SinOsc.ar/0:
                    frequency: 333.0
                    phase: 0.0
            -   Pan2.ar/0:
                    level: 1.0
                    position: -1.0
                    source: SinOsc.ar/0[0]
            -   SinOsc.ar/1:
                    frequency: 444.0
                    phase: 0.0
            -   Pan2.ar/1:
                    level: 1.0
                    position: -0.5
                    source: SinOsc.ar/1[0]
            -   SinOsc.ar/2:
                    frequency: 555.0
                    phase: 0.0
            -   Pan2.ar/2:
                    level: 1.0
                    position: 0.0
                    source: SinOsc.ar/2[0]
            -   SinOsc.ar/3:
                    frequency: 666.0
                    phase: 0.0
            -   Pan2.ar/3:
                    level: 1.0
                    position: 0.5
                    source: SinOsc.ar/3[0]
            -   Sum4.ar/0:
                    input_four: Pan2.ar/3[0]
                    input_one: Pan2.ar/0[0]
                    input_three: Pan2.ar/2[0]
                    input_two: Pan2.ar/1[0]
            -   Sum4.ar/1:
                    input_four: Pan2.ar/3[1]
                    input_one: Pan2.ar/0[1]
                    input_three: Pan2.ar/2[1]
                    input_two: Pan2.ar/1[1]
            -   SinOsc.ar/4:
                    frequency: 777.0
                    phase: 0.0
            -   Pan2.ar/4:
                    level: 1.0
                    position: 1.0
                    source: SinOsc.ar/4[0]
            -   BinaryOpUGen(ADDITION).ar/0:
                    left: Sum4.ar/0[0]
                    right: Pan2.ar/4[0]
            -   BinaryOpUGen(MULTIPLICATION).ar/0:
                    left: BinaryOpUGen(ADDITION).ar/0[0]
                    right: 0.4472135954999579
            -   BinaryOpUGen(ADDITION).ar/1:
                    left: Sum4.ar/1[0]
                    right: Pan2.ar/4[1]
            -   BinaryOpUGen(MULTIPLICATION).ar/1:
                    left: BinaryOpUGen(ADDITION).ar/1[0]
                    right: 0.4472135954999579

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [
            ("spread", 1),
            ("level", 1),
            ("center", 0),
            ("normalize", True),
            ("source", None),
        ]
    )
    _unexpanded_input_names = ("source",)

    @classmethod
    def _new_expanded(cls, calculation_rate=None, **kwargs):
        dictionaries = UGen._expand_dictionary(
            kwargs, unexpanded_input_names=["source"]
        )
        ugens = [
            cls._new_single(calculation_rate=calculation_rate, **dictionary)
            for dictionary in dictionaries
        ]
        return Mix.multichannel(ugens, 2)

    ### CLASS METHODS ###

    @classmethod
    def _new_single(
        cls,
        calculation_rate=None,
        center=0,
        level=1,
        normalize=True,
        source=None,
        spread=1,
    ):
        positions = [
            (i * (2 / (len(source) - 1)) - 1) * spread + center
            for i in range(len(source))
        ]
        if normalize:
            if calculation_rate == CalculationRate.AUDIO:
                level = level * math.sqrt(1 / len(source))
            else:
                level = level / len(source)
        panners = UGen._get_method_for_rate(Pan2, calculation_rate)(
            source=source, position=positions
        )
        return Mix.multichannel(panners, 2) * level

    @classmethod
    def ar(cls, *, center=0, level=1, normalize=True, source=None, spread=1):
        return cls._new_expanded(
            calculation_rate=CalculationRate.AUDIO,
            center=center,
            level=level,
            normalize=normalize,
            source=source,
            spread=spread,
        )

    @classmethod
    def kr(cls, *, center=0, level=1, normalize=True, source=None, spread=1):
        return cls._new_expanded(
            calculation_rate=CalculationRate.CONTROL,
            center=center,
            level=level,
            normalize=normalize,
            source=source,
            spread=spread,
        )


class XFade2(UGen):
    """
    Two channel equal power crossfader.

    ::

        >>> xfade_3 = supriya.ugens.XFade2.ar(
        ...     in_a=supriya.ugens.Saw.ar(),
        ...     in_b=supriya.ugens.SinOsc.ar(),
        ...     level=1,
        ...     pan=supriya.ugens.LFTri.kr(frequency=0.1),
        ... )
        >>> xfade_3
        XFade2.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("in_a", None), ("in_b", 0), ("pan", 0), ("level", 1)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



from collections.abc import Sequence
from typing import NamedTuple, Tuple, Union

import supriya.osc
from supriya.enums import NodeAction, RequestId

from .bases import Request, Response


class NodeFreeRequest(Request):
    """
    A /n_free request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.NodeFreeRequest(
        ...     node_ids=1000,
        ... )
        >>> request
        NodeFreeRequest(
            node_ids=(1000,),
        )

    ::

        >>> request.to_osc()
        OscMessage('/n_free', 1000)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.NODE_FREE

    ### INITIALIZER ###

    def __init__(self, node_ids=None):
        Request.__init__(self)
        if not isinstance(node_ids, Sequence):
            node_ids = (node_ids,)
        node_ids = tuple(int(_) for _ in node_ids)
        self._node_ids = node_ids

    ### PRIVATE METHODS ###

    def _apply_local(self, server):
        for node_id in self.node_ids:
            node = server._nodes.get(node_id)
            if not node:
                continue
            node._set_parent(None)
            node._unregister_with_local_server()

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id, *self.node_ids]
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def node_ids(self):
        return self._node_ids

    @property
    def response_patterns(self):
        return ["/n_end", int(self.node_ids[-1])], None


class NodeInfoResponse(Response):

    ### INITIALIZER ###

    def __init__(
        self,
        action=None,
        node_id=None,
        parent_id=None,
        previous_node_id=None,
        next_node_id=None,
        is_group=None,
        head_node_id=None,
        tail_node_id=None,
        synthdef_name=None,
        synthdef_controls=None,
    ):
        self._action = NodeAction.from_address(action)
        self._is_group = bool(is_group)
        self._head_node_id = self._coerce_node_id(head_node_id)
        self._next_node_id = self._coerce_node_id(next_node_id)
        self._node_id = self._coerce_node_id(node_id)
        self._parent_id = self._coerce_node_id(parent_id)
        self._previous_node_id = self._coerce_node_id(previous_node_id)
        self._tail_node_id = self._coerce_node_id(tail_node_id)
        self._synthdef_name = synthdef_name
        self._synthdef_controls = synthdef_controls

    ### PRIVATE METHODS ###

    def _coerce_node_id(self, node_id):
        if node_id is not None and -1 < node_id:
            return node_id
        return None

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        arguments = (osc_message.address,) + osc_message.contents
        kwargs = dict(
            action=arguments[0],
            node_id=arguments[1],
            parent_id=arguments[2],
            previous_node_id=arguments[3],
            next_node_id=arguments[4],
            is_group=arguments[5],
        )
        if arguments[5]:
            kwargs.update(head_node_id=arguments[6], tail_node_id=arguments[7])
        elif len(arguments) > 6:
            controls = []
            for i in range(arguments[7]):
                controls.append((arguments[8 + (i * 2)], arguments[9 + (i * 2)]))
            kwargs.update(synthdef_name=arguments[6], synthdef_controls=tuple(controls))
        response = cls(**kwargs)
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def action(self):
        return self._action

    @property
    def head_node_id(self):
        return self._head_node_id

    @property
    def is_group(self):
        return self._is_group

    @property
    def next_node_id(self):
        return self._next_node_id

    @property
    def node_id(self):
        return self._node_id

    @property
    def parent_id(self):
        return self._parent_id

    @property
    def previous_node_id(self):
        return self._previous_node_id

    @property
    def tail_node_id(self):
        return self._tail_node_id

    @property
    def synthdef_controls(self):
        return self._synthdef_controls

    @property
    def synthdef_name(self):
        return self._synthdef_name


class NodeMapToAudioBusRequest(Request):
    """
    A /n_mapa request.

    ::

        >>> import supriya.commands
        >>> import supriya.realtime
        >>> request = supriya.commands.NodeMapToAudioBusRequest(
        ...     node_id=1000,
        ...     frequency=supriya.realtime.Bus(9, "audio"),
        ...     phase=supriya.realtime.Bus(10, "audio"),
        ...     amplitude=supriya.realtime.Bus(11, "audio"),
        ... )
        >>> request
        NodeMapToAudioBusRequest(
            amplitude=<- Bus: 11 (audio)>,
            frequency=<- Bus: 9 (audio)>,
            node_id=1000,
            phase=<- Bus: 10 (audio)>,
        )

    ::

        >>> request.to_osc()
        OscMessage('/n_mapa', 1000, 'amplitude', 11, 'frequency', 9, 'phase', 10)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.NODE_MAP_TO_AUDIO_BUS

    ### INITIALIZER ###

    def __init__(self, node_id=None, **kwargs):
        Request.__init__(self)
        self._node_id = node_id
        self._kwargs = dict((name, value) for name, value in kwargs.items())

    ### SPECIAL METHODS ###

    def __getattr__(self, name):
        if name in self._kwargs:
            return self._kwargs[name]
        return object.__getattr__(self, name)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        node_id = self._sanitize_node_id(self.node_id, with_placeholders)
        contents = []
        for name, bus in sorted(self._kwargs.items()):
            contents.append(name)
            contents.append(int(bus))
        message = supriya.osc.OscMessage(request_id, node_id, *contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def node_id(self):
        return self._node_id


class NodeMapToControlBusRequest(Request):
    """
    A /n_map request.

    ::

        >>> import supriya.commands
        >>> import supriya.realtime
        >>> request = supriya.commands.NodeMapToControlBusRequest(
        ...     node_id=1000,
        ...     frequency=supriya.realtime.Bus(9, "control"),
        ...     phase=supriya.realtime.Bus(10, "control"),
        ...     amplitude=supriya.realtime.Bus(11, "control"),
        ... )
        >>> request
        NodeMapToControlBusRequest(
            amplitude=<- Bus: 11 (control)>,
            frequency=<- Bus: 9 (control)>,
            node_id=1000,
            phase=<- Bus: 10 (control)>,
        )

    ::

        >>> request.to_osc()
        OscMessage('/n_map', 1000, 'amplitude', 11, 'frequency', 9, 'phase', 10)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.NODE_MAP_TO_CONTROL_BUS

    ### INITIALIZER ###

    def __init__(self, node_id=None, **kwargs):
        Request.__init__(self)
        self._node_id = node_id
        self._kwargs = dict((name, value) for name, value in kwargs.items())

    ### SPECIAL METHODS ###

    def __getattr__(self, name):
        if name in self._kwargs:
            return self._kwargs[name]
        return object.__getattr__(self, name)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        node_id = self._sanitize_node_id(self.node_id, with_placeholders)
        contents = []
        for name, bus in sorted(self._kwargs.items()):
            contents.append(name)
            contents.append(int(bus))
        message = supriya.osc.OscMessage(request_id, node_id, *contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def node_id(self):
        return self._node_id


class NodeQueryRequest(Request):
    """
    A /n_query request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.NodeQueryRequest(
        ...     node_id=1000,
        ... )
        >>> request
        NodeQueryRequest(
            node_id=1000,
        )

    ::

        >>> request.to_osc()
        OscMessage('/n_query', 1000)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.NODE_QUERY

    ### INITIALIZER ###

    def __init__(self, node_id=None):
        Request.__init__(self)
        self._node_id = node_id

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        node_id = int(self.node_id)
        message = supriya.osc.OscMessage(request_id, node_id)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def node_id(self):
        return self._node_id

    @property
    def response_patterns(self):
        return ["/n_info", self.node_id], ["/fail"]


class NodeRunRequest(Request):
    """
    A /n_run request.

    ::

        >>> server = supriya.Server().boot()
        >>> synth_a = supriya.Synth().allocate(server)
        >>> synth_b = supriya.Synth().allocate(server)
        >>> synth_a.is_paused, synth_b.is_paused
        (False, False)

    Unpause ``synth_a`` (a no-op because it's already unpaused) and pause
    ``synth_b``:

    ::

        >>> request = supriya.commands.NodeRunRequest(
        ...     [
        ...         [synth_a, True],
        ...         [synth_b, False],
        ...     ]
        ... )
        >>> request.to_osc()
        OscMessage('/n_run', 1000, 1, 1001, 0)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     request.communicate(server=server)
        ...     _ = server.sync()
        ...
        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/n_run', 1000, 1, 1001, 0))
        ('S', OscMessage('/sync', 0))
        ('R', OscMessage('/n_off', 1001, 1, -1, 1000, 0))
        ('R', OscMessage('/synced', 0))

    ::

        >>> synth_a.is_paused, synth_b.is_paused
        (False, True)

    Pause ``synth_a`` and unpause ``synth_b``:

    ::

        >>> request = supriya.commands.NodeRunRequest(
        ...     [
        ...         [synth_a, False],
        ...         [synth_b, True],
        ...     ]
        ... )
        >>> request.to_osc()
        OscMessage('/n_run', 1000, 0, 1001, 1)

    ::

        >>> with server.osc_protocol.capture() as transcript:
        ...     request.communicate(server=server)
        ...     _ = server.sync()
        ...
        >>> for entry in transcript:
        ...     (entry.label, entry.message)
        ...
        ('S', OscMessage('/n_run', 1000, 0, 1001, 1))
        ('S', OscMessage('/sync', 1))
        ('R', OscMessage('/n_off', 1000, 1, 1001, -1, 0))
        ('R', OscMessage('/n_on', 1001, 1, -1, 1000, 0))
        ('R', OscMessage('/synced', 1))

    ::

        >>> synth_a.is_paused, synth_b.is_paused
        (True, False)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.NODE_RUN

    ### INITIALIZER ###

    def __init__(self, node_id_run_flag_pairs=None):
        Request.__init__(self)
        if node_id_run_flag_pairs:
            pairs = []
            for node_id, run_flag in node_id_run_flag_pairs:
                node_id = node_id
                run_flag = bool(run_flag)
                pairs.append((node_id, run_flag))
            node_id_run_flag_pairs = tuple(pairs)
        self._node_id_run_flag_pairs = node_id_run_flag_pairs

    ### PRIVATE METHODS ###

    def _apply_local(self, server):
        for node_id, run_flag in self.node_id_run_flag_pairs:
            node = server._nodes.get(node_id)
            if not node:
                continue
            node._run(run_flag)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id]
        sanitized_pairs = []
        for node_id, run_flag in self.node_id_run_flag_pairs or []:
            node_id = self._sanitize_node_id(node_id, with_placeholders)
            sanitized_pairs.append((node_id, int(run_flag)))
        for pair in sorted(sanitized_pairs):
            contents.extend(pair)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def node_id_run_flag_pairs(self):
        return self._node_id_run_flag_pairs


class NodeSetRequest(Request):
    """
    A /n_set request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.NodeSetRequest(
        ...     1000,
        ...     frequency=443.1,
        ...     phase=0.5,
        ...     amplitude=0.1,
        ... )
        >>> request
        NodeSetRequest(
            amplitude=0.1,
            frequency=443.1,
            node_id=1000,
            phase=0.5,
        )

    ::

        >>> request.to_osc()
        OscMessage('/n_set', 1000, 'amplitude', 0.1, 'frequency', 443.1, 'phase', 0.5)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.NODE_SET

    ### INITIALIZER ###

    def __init__(self, node_id=None, **kwargs):
        Request.__init__(self)
        self._node_id = node_id
        self._kwargs = kwargs

    ### SPECIAL METHODS ###

    def __getattr__(self, name):
        if name in self._kwargs:
            return self._kwargs[name]
        return object.__getattr__(self, name)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        node_id = int(self.node_id)
        contents = [request_id, node_id]
        for key, value in sorted(self._kwargs.items()):
            contents.append(key)
            contents.append(value)
        message = supriya.osc.OscMessage(*contents)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def node_id(self):
        return self._node_id


class NodeSetResponse(Response):

    ### CLASS VARIABLES ###

    class Item(NamedTuple):
        control_index_or_name: Union[int, str]
        control_value: float

    ### INITIALIZER ###

    def __init__(self, node_id=None, items=None):
        self._items = items
        self._node_id = node_id

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        """
        Create response from OSC message.

        ::

            >>> message = supriya.osc.OscMessage("/n_set", 1023, "/one", -1, "/two", 0)
            >>> supriya.commands.NodeSetResponse.from_osc_message(message)
            NodeSetResponse(
                items=(
                    Item(control_index_or_name='/one', control_value=-1),
                    Item(control_index_or_name='/two', control_value=0),
                ),
                node_id=1023,
            )

        """
        node_id, remainder = osc_message.contents[0], osc_message.contents[1:]
        items = []
        for group in cls._group_items(remainder, 2):
            item = cls.Item(*group)
            items.append(item)
        response = cls(node_id=node_id, items=tuple(items))
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def items(self):
        return self._items

    @property
    def node_id(self):
        return self._node_id


class NodeSetContiguousResponse(Response):

    ### CLASS VARIABLES ###

    class Item(NamedTuple):
        control_values: Tuple[float]
        starting_control_index_or_name: Union[int, str]

    ### INITIALIZER ###

    def __init__(self, node_id=None, items=None):
        self._items = items
        self._node_id = node_id

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        node_id, remainder = osc_message.contents[0], osc_message.contents[1:]
        items = []
        while remainder:
            control_index_or_name = remainder[0]
            control_count = remainder[1]
            control_values = tuple(remainder[2 : 2 + control_count])
            item = cls.Item(
                control_index_or_name=control_index_or_name,
                control_values=control_values,
            )
            items.append(item)
            remainder = remainder[2 + control_count :]
        items = tuple(items)
        response = cls(node_id=node_id, items=items)
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def items(self):
        return self._items

    @property
    def node_id(self):
        return self._node_id



"""
The core pattern classes.
"""
import abc
import inspect
import itertools
import operator
import random
from collections.abc import Sequence
from typing import Dict, Iterator, Optional

from uqbar.objects import get_vars

from supriya.clocks import Clock, OfflineClock
from supriya.providers import NonrealtimeProvider, Provider

from .events import CompositeEvent
from .players import PatternPlayer


class Pattern(metaclass=abc.ABCMeta):

    ### CLASSMETHODS ###

    _rngs: Dict[int, Iterator[float]] = {}

    ### SPECIAL METHODS ###

    def __abs__(self):
        return UnaryOpPattern("abs", self)

    def __add__(self, expr):
        return BinaryOpPattern("+", self, expr)

    def __eq__(self, expr):
        self_values = type(self), get_vars(self)
        try:
            expr_values = type(expr), get_vars(expr)
        except AttributeError:
            expr_values = type(expr), expr
        return self_values == expr_values

    def __floordiv__(self, expr):
        return BinaryOpPattern("//", self, expr)

    def __invert__(self):
        return UnaryOpPattern("~", self)

    def __iter__(self):
        should_stop = False
        state: Optional[Dict] = self._setup_state()
        iterator = self._iterate(state)
        try:
            expr = self._adjust_recursive(next(iterator), state=state)
        except StopIteration:
            return
        start_event, stop_event = self._setup_peripherals(state)
        if start_event:
            should_stop = (yield start_event) or should_stop
        if not should_stop:
            should_stop = (yield expr) or should_stop
            while True:  # Exhaust iterator, even if scheduled to stop
                try:
                    expr = self._adjust_recursive(
                        iterator.send(should_stop), state=state
                    )
                    should_stop = (yield expr) or should_stop
                except StopIteration:
                    break
        if stop_event:
            yield stop_event

    def __mod__(self, expr):
        return BinaryOpPattern("%", self, expr)

    def __mul__(self, expr):
        return BinaryOpPattern("*", self, expr)

    def __neg__(self):
        return UnaryOpPattern("-", self)

    def __pos__(self):
        return UnaryOpPattern("+", self)

    def __pow__(self, expr):
        return BinaryOpPattern("**", self, expr)

    def __radd__(self, expr):
        return BinaryOpPattern("+", expr, self)

    def __rmod__(self, expr):
        return BinaryOpPattern("%", expr, self)

    def __rmul__(self, expr):
        return BinaryOpPattern("*", expr, self)

    def __rpow__(self, expr):
        return BinaryOpPattern("**", expr, self)

    def __rsub__(self, expr):
        return BinaryOpPattern("-", expr, self)

    def __rtruediv__(self, expr):
        return BinaryOpPattern("/", expr, self)

    def __rfloordiv__(self, expr):
        return BinaryOpPattern("//", expr, self)

    def __sub__(self, expr):
        return BinaryOpPattern("-", self, expr)

    def __truediv__(self, expr):
        return BinaryOpPattern("/", self, expr)

    ### PRIVATE METHODS ###

    def _adjust(self, expr, state=None):
        return expr

    def _adjust_recursive(self, expr, state=None):
        if isinstance(expr, CompositeEvent):
            return CompositeEvent(
                [self._adjust(event, state=state) for event in expr.events],
                delta=expr.delta,
            )
        return self._adjust(expr, state=state)

    def _apply_recursive(self, procedure, *exprs):
        if all(not isinstance(x, Sequence) for x in exprs):
            return procedure(*exprs)
        coerced_exprs = [
            expr if isinstance(expr, Sequence) else [expr] for expr in exprs
        ]
        max_length = max(len(expr) for expr in coerced_exprs)
        for i, expr in enumerate(coerced_exprs):
            if len(expr) < max_length:
                cycle = itertools.cycle(expr)
                coerced_exprs[i] = [next(cycle) for _ in range(max_length)]
        return tuple(
            self._apply_recursive(procedure, *items) for items in zip(*coerced_exprs)
        )

    def _freeze_recursive(self, value):
        if isinstance(value, str):
            return value
        elif isinstance(value, Sequence) and not isinstance(value, Pattern):
            return tuple(self._freeze_recursive(_) for _ in value)
        return value

    def _get_rng(self):
        identifier = None
        try:
            # Walk frames to find an enclosing SeedPattern._iterate()
            frame = inspect.currentframe()
            while frame is not None:
                if (
                    isinstance(frame.f_locals.get("self"), SeedPattern)
                    and frame.f_code.co_name == "_iterate"
                ):
                    identifier = id(frame)
                    break
                frame = frame.f_back
        finally:
            del frame
        if identifier in self._rngs:
            return self._rngs[identifier]
        return self._get_stdlib_rng()

    def _get_seeded_rng(self, seed: int = 1) -> Iterator[float]:
        while True:
            seed = (seed * 1_103_515_245 + 12345) & 0x7FFFFFFF
            yield float(seed) / 0x7FFFFFFF

    def _get_stdlib_rng(self) -> Iterator[float]:
        while True:
            yield random.random()

    @abc.abstractmethod
    def _iterate(self):
        raise NotImplementedError

    def _loop(self, iterations=None):
        if iterations is None:
            while True:
                yield True
        else:
            for _ in range(iterations):
                yield True

    def _setup_state(self) -> Optional[Dict]:
        return None

    def _setup_peripherals(self, state):
        return None, None

    ### PUBLIC METHODS ###

    def play(
        self,
        provider=None,
        *,
        at=None,
        clock=None,
        quantization=None,
        tempo=None,
        until=None,
    ):
        if provider is None:
            provider = Provider.realtime()
        elif not isinstance(provider, Provider):
            provider = Provider.from_context(provider)
        if isinstance(provider, NonrealtimeProvider):
            clock = OfflineClock()
            at = at or 0.0
        elif not clock:
            clock = Clock.default()
        player = PatternPlayer(pattern=self, provider=provider, clock=clock)
        player.play(quantization=quantization, at=at, until=until)
        return player

    ### PUBLIC PROPERTIES ###

    @abc.abstractproperty
    def is_infinite(self):
        raise NotImplementedError


class BinaryOpPattern(Pattern):

    ### INITIALIZER ###

    def __init__(self, operator, expr_one, expr_two):
        self._operator = operator
        self._expr_one = self._freeze_recursive(expr_one)
        self._expr_two = self._freeze_recursive(expr_two)

    ### PRIVATE METHODS ###

    def _iterate(self, state=None):
        expr_one = self.expr_one
        if not isinstance(expr_one, Pattern):
            expr_one = SequencePattern([expr_one], None)
        expr_one = iter(expr_one)
        expr_two = self.expr_two
        if not isinstance(expr_two, Pattern):
            expr_two = SequencePattern([expr_two], None)
        expr_two = iter(expr_two)
        operator = self._string_to_operator()
        for item_one, item_two in zip(expr_one, expr_two):
            yield self._apply_recursive(operator, item_one, item_two)

    def _string_to_operator(self):
        operators = {
            "%": operator.__mod__,
            "*": operator.__mul__,
            "**": operator.__pow__,
            "+": operator.__add__,
            "-": operator.__sub__,
            "/": operator.__truediv__,
            "//": operator.__floordiv__,
        }
        return operators[self.operator]

    ### PUBLIC PROPERTIES ###

    @property
    def expr_one(self):
        return self._expr_one

    @property
    def expr_two(self):
        return self._expr_two

    @property
    def is_infinite(self):
        expr_one_is_infinite = (
            not isinstance(self.expr_one, Pattern) or self.expr_one.is_infinite
        )
        expr_two_is_infinite = (
            not isinstance(self.expr_two, Pattern) or self.expr_two.is_infinite
        )
        return expr_one_is_infinite and expr_two_is_infinite

    @property
    def operator(self):
        return self._operator


class UnaryOpPattern(Pattern):

    ### INITIALIZER ###

    def __init__(self, operator, expr):
        self._operator = operator
        self._expr = expr

    ### PRIVATE METHODS ###

    def _iterate(self, state=None):
        expr = self.expr
        if not isinstance(expr, Pattern):
            expr = SequencePattern([expr], None)
        expr = iter(expr)
        operator = self._string_to_operator()
        for item in expr:
            yield self._apply_recursive(operator, item)

    def _string_to_operator(self):
        operators = {
            "~": operator.invert,
            "-": operator.__neg__,
            "+": operator.__pos__,
            "abs": operator.abs,
        }
        return operators[self.operator]

    ### PUBLIC PROPERTIES ###

    @property
    def expr(self):
        return self._expr

    @property
    def is_infinite(self):
        return not isinstance(self.expr, Pattern) or self.expr.is_infinite

    @property
    def operator(self):
        return self._operator


class SeedPattern(Pattern):

    ### INITIALIZER ###

    def __init__(self, pattern, seed=0):
        if not isinstance(pattern, Pattern):
            raise ValueError(f"Must be pattern: {pattern!r}")
        self._pattern = pattern
        self._seed = int(seed)

    ### PRIVATE METHODS ###

    def _iterate(self, state=None):
        try:
            identifier = id(inspect.currentframe())
            rng = self._get_seeded_rng(seed=self.seed)
            self._rngs[identifier] = rng
            yield from self._pattern
        finally:
            del self._rngs[identifier]

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self):
        return self._pattern.is_infinite

    @property
    def pattern(self):
        return self._pattern

    @property
    def seed(self):
        return self._seed


class SequencePattern(Pattern):

    ### INITIALIZER ###

    def __init__(self, sequence, iterations=1):
        if not isinstance(sequence, Sequence):
            raise ValueError(f"Must be sequence: {sequence!r}")
        if iterations is not None:
            iterations = int(iterations)
            if iterations < 1:
                raise ValueError("Iterations must be null or greater than 0")
        self._sequence = self._freeze_recursive(sequence)
        self._iterations = iterations

    ### PRIVATE METHODS ###

    def _iterate(self, state=None):
        should_stop = False
        for _ in self._loop(self._iterations):
            for x in self._sequence:
                if not isinstance(x, Pattern):
                    should_stop = (yield x) or should_stop
                else:
                    iterator = iter(x)
                    try:
                        y = next(iterator)
                        should_stop = (yield y) or should_stop
                        while True:
                            y = iterator.send(should_stop)
                            should_stop = (yield y) or should_stop
                    except StopIteration:
                        pass
                if should_stop:
                    return

    ### PUBLIC PROPERTIES ###

    @property
    def is_infinite(self):
        if self._iterations is None:
            return True
        for x in self._sequence:
            if isinstance(x, Pattern) and x.is_infinite:
                return True
        return False



import abc
import fnmatch
import os
from typing import List, Union

import uqbar.objects

import supriya


def _search(pattern: str, root_path: str):
    search_path, pattern = os.path.split(pattern)
    search_path = os.path.expanduser(search_path)
    if not search_path:
        search_path = os.path.join(root_path, "assets")
    elif not os.path.isabs(search_path):
        search_path = os.path.join(root_path, "assets", search_path)
    result: List[str] = []
    result = os.listdir(search_path)
    result = fnmatch.filter(result, pattern)
    result = [os.path.join(search_path, _) for _ in result]
    if len(result) == 1:
        return result[0]
    return result


class _AssetsMeta(abc.ABCMeta):

    root_path: str = supriya.__path__[0]  # type: ignore

    def __getitem__(self, pattern) -> Union[str, List[str]]:
        return _search(pattern, self.root_path)


class Assets(metaclass=_AssetsMeta):
    def __init__(self, root_path: str) -> None:
        self.root_path = root_path

    def __getitem__(self, pattern) -> Union[str, List[str]]:
        return _search(pattern, self.root_path)


class SupriyaObject(metaclass=abc.ABCMeta):
    """
    Abstract base class from which many custom classes inherit.
    """

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __repr__(self):
        return uqbar.objects.get_repr(self, multiline=True)


class SupriyaValueObject(SupriyaObject):

    ### CLASS VARIABLES ###

    __slots__ = ()

    ### SPECIAL METHODS ###

    def __copy__(self, *args):
        return uqbar.objects.new(self)

    def __eq__(self, expr):
        self_values = type(self), uqbar.objects.get_vars(self)
        try:
            expr_values = type(expr), uqbar.objects.get_vars(expr)
        except AttributeError:
            expr_values = type(expr), expr
        return self_values == expr_values

    def __hash__(self):
        args, var_args, kwargs = uqbar.objects.get_vars(self)
        hash_values = [type(self)]
        hash_values.append(tuple(args.items()))
        hash_values.append(tuple(var_args))
        hash_values.append(tuple(sorted(kwargs.items())))
        return hash(tuple(hash_values))



import supriya.osc
from supriya.enums import RequestId

from .bases import Request, Response


class ClearScheduleRequest(Request):

    ### CLASS VARIABLES ###

    request_id = RequestId.CLEAR_SCHEDULE

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        contents = [request_id]
        message = supriya.osc.OscMessage(*contents)
        return message


class DoneResponse(Response):

    ### INITIALIZER ###

    def __init__(self, action=None):
        self._action = action

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        arguments = osc_message.contents
        response = cls(action=tuple(arguments))
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def action(self):
        return self._action


class DumpOscRequest(Request):
    """
    A /dumpOSC request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.DumpOscRequest(1)
        >>> request
        DumpOscRequest(
            osc_status=1,
        )

    ::

        >>> request.to_osc()
        OscMessage('/dumpOSC', 1)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.DUMP_OSC

    ### INITIALIZER ###

    def __init__(self, osc_status=None):
        Request.__init__(self)
        self._osc_status = int(osc_status)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        osc_status = int(self.osc_status)
        assert 0 <= osc_status <= 4
        message = supriya.osc.OscMessage(request_id, osc_status)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def osc_status(self):
        return self._osc_status


class FailResponse(Response):

    ### INITIALIZER ###

    def __init__(self, failed_command=None, failure_reason=None):
        self._failed_command = failed_command
        self._failure_reason = failure_reason

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        failed_command = osc_message.contents[0]
        failure_reason = osc_message.contents[1:]
        if failure_reason:
            failure_reason = tuple(failure_reason)
        response = cls(failed_command=failed_command, failure_reason=failure_reason)
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def failed_command(self):
        return self._failed_command

    @property
    def failure_reason(self):
        return self._failure_reason


class NothingRequest(Request):

    ### CLASS VARIABLES ###

    request_id = RequestId.NOTHING

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        return supriya.osc.OscMessage(0)


class NotifyRequest(Request):
    """
    A /notify message.

    ::

        >>> import supriya.commands

        >>> request = supriya.commands.NotifyRequest(
        ...     notify_status=True,
        ... )
        >>> request
        NotifyRequest(
            notify_status=True,
        )

    ::

        >>> request.to_osc()
        OscMessage('/notify', 1)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.NOTIFY

    ### INITIALIZER ###

    def __init__(self, notify_status=None):
        Request.__init__(self)
        self._notify_status = bool(notify_status)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        notify_status = int(self.notify_status)
        message = supriya.osc.OscMessage(request_id, notify_status)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def notify_status(self):
        return self._notify_status

    @property
    def response_patterns(self):
        return ["/done", "/notify"], ["/fail", "/notify"]


class QuitRequest(Request):

    ### CLASS VARIABLES ###

    request_id = RequestId.QUIT

    ### INITIALIZER ###

    def __init__(self):
        Request.__init__(self)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        message = supriya.osc.OscMessage(request_id)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def response_patterns(self):
        return ["/done", "/quit"], None


class StatusRequest(Request):
    """
    A /status request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.StatusRequest()
        >>> request
        StatusRequest()

    ::

        >>> request.to_osc()
        OscMessage('/status')

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.STATUS

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        message = supriya.osc.OscMessage(request_id)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def response_patterns(self):
        return ["/status.reply"], None


class StatusResponse(Response):

    ### INITIALIZER ###

    def __init__(
        self,
        actual_sample_rate=None,
        average_cpu_usage=None,
        group_count=None,
        peak_cpu_usage=None,
        synth_count=None,
        synthdef_count=None,
        target_sample_rate=None,
        ugen_count=None,
    ):
        self._actual_sample_rate = actual_sample_rate
        self._average_cpu_usage = average_cpu_usage
        self._group_count = group_count
        self._peak_cpu_usage = peak_cpu_usage
        self._synth_count = synth_count
        self._synthdef_count = synthdef_count
        self._target_sample_rate = target_sample_rate
        self._ugen_count = ugen_count

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        """
        Create response from OSC message.

        ::

            >>> message = supriya.osc.OscMessage(
            ...     "/status.reply",
            ...     1,
            ...     0,
            ...     0,
            ...     2,
            ...     4,
            ...     0.040679048746824265,
            ...     0.15118031203746796,
            ...     44100.0,
            ...     44100.00077873274,
            ... )
            >>> supriya.commands.StatusResponse.from_osc_message(message)
            StatusResponse(
                actual_sample_rate=44100.00077873274,
                average_cpu_usage=0.040679048746824265,
                group_count=2,
                peak_cpu_usage=0.15118031203746796,
                synth_count=0,
                synthdef_count=4,
                target_sample_rate=44100.0,
                ugen_count=0,
            )

        """
        arguments = osc_message.contents[1:]
        (
            ugen_count,
            synth_count,
            group_count,
            synthdef_count,
            average_cpu_usage,
            peak_cpu_usage,
            target_sample_rate,
            actual_sample_rate,
        ) = arguments
        response = cls(
            actual_sample_rate=actual_sample_rate,
            average_cpu_usage=average_cpu_usage,
            group_count=group_count,
            peak_cpu_usage=peak_cpu_usage,
            synth_count=synth_count,
            synthdef_count=synthdef_count,
            target_sample_rate=target_sample_rate,
            ugen_count=ugen_count,
        )
        return response

    def to_dict(self):
        """
        Convert StatusResponse to JSON-serializable dictionay.

        ::

            >>> status_response = supriya.commands.StatusResponse(
            ...     actual_sample_rate=44100.05692801021,
            ...     average_cpu_usage=8.151924133300781,
            ...     group_count=6,
            ...     peak_cpu_usage=15.151398658752441,
            ...     synth_count=19,
            ...     synthdef_count=42,
            ...     target_sample_rate=44100.0,
            ...     ugen_count=685,
            ... )

        ::

            >>> import json
            >>> result = status_response.to_dict()
            >>> result = json.dumps(
            ...     result,
            ...     indent=4,
            ...     separators=(",", ": "),
            ...     sort_keys=True,
            ... )
            >>> print(result)
            {
                "server_status": {
                    "actual_sample_rate": 44100.05692801021,
                    "average_cpu_usage": 8.151924133300781,
                    "group_count": 6,
                    "peak_cpu_usage": 15.151398658752441,
                    "synth_count": 19,
                    "synthdef_count": 42,
                    "target_sample_rate": 44100.0,
                    "ugen_count": 685
                }
            }

        """
        result = {
            "server_status": {
                "actual_sample_rate": self.actual_sample_rate,
                "average_cpu_usage": self.average_cpu_usage,
                "group_count": self.group_count,
                "peak_cpu_usage": self.peak_cpu_usage,
                "synth_count": self.synth_count,
                "synthdef_count": self.synthdef_count,
                "target_sample_rate": self.target_sample_rate,
                "ugen_count": self.ugen_count,
            }
        }
        return result

    ### PUBLIC PROPERTIES ###

    @property
    def actual_sample_rate(self):
        return self._actual_sample_rate

    @property
    def average_cpu_usage(self):
        return self._average_cpu_usage

    @property
    def group_count(self):
        return self._group_count

    @property
    def peak_cpu_usage(self):
        return self._peak_cpu_usage

    @property
    def synth_count(self):
        return self._synth_count

    @property
    def synthdef_count(self):
        return self._synthdef_count

    @property
    def target_sample_rate(self):
        return self._target_sample_rate

    @property
    def ugen_count(self):
        return self._ugen_count


class SyncedResponse(Response):

    ### INITIALIZER ###

    def __init__(self, sync_id=None):
        self._sync_id = sync_id

    ### PUBLIC METHODS ###

    @classmethod
    def from_osc_message(cls, osc_message):
        arguments = osc_message.contents
        response = cls(*arguments)
        return response

    ### PUBLIC PROPERTIES ###

    @property
    def sync_id(self):
        return self._sync_id


class SyncRequest(Request):
    """
    A /sync request.

    ::

        >>> import supriya.commands
        >>> request = supriya.commands.SyncRequest(
        ...     sync_id=1999,
        ... )
        >>> request
        SyncRequest(
            sync_id=1999,
        )

    ::

        >>> request.to_osc()
        OscMessage('/sync', 1999)

    """

    ### CLASS VARIABLES ###

    request_id = RequestId.SYNC

    ### INITIALIZER ###

    def __init__(self, sync_id=None):
        Request.__init__(self)
        self._sync_id = int(sync_id)

    ### PUBLIC METHODS ###

    def to_osc(self, *, with_placeholders=False):
        request_id = self.request_name
        sync_id = int(self.sync_id)
        message = supriya.osc.OscMessage(request_id, sync_id)
        return message

    ### PUBLIC PROPERTIES ###

    @property
    def response_patterns(self):
        return ["/synced", self.sync_id], None

    @property
    def sync_id(self):
        return self._sync_id



import collections

from supriya import CalculationRate, SignalRange
from supriya.synthdefs import UGen, WidthFirstUGen
from supriya.typing import UGenInputMap


class BrownNoise(UGen):
    """
    A brown noise unit generator.

    ::

        >>> supriya.ugens.BrownNoise.ar()
        BrownNoise.ar()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class ClipNoise(UGen):
    """
    A clipped noise unit generator.

    ::

        >>> supriya.ugens.ClipNoise.ar()
        ClipNoise.ar()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class CoinGate(UGen):
    """
    A probabilistic trigger gate.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> coin_gate = supriya.ugens.CoinGate.ar(
        ...     probability=0.5,
        ...     trigger=trigger,
        ... )
        >>> coin_gate
        CoinGate.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("probability", 0.5), ("trigger", None)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Crackle(UGen):
    """
    A chaotic noise generator.

    ::

        >>> crackle = supriya.ugens.Crackle.ar(
        ...     chaos_parameter=1.25,
        ... )
        >>> crackle
        Crackle.ar()

    """

    _ordered_input_names = collections.OrderedDict([("chaos_parameter", 1.5)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Dust(UGen):
    """
    A unipolar random impulse generator.

    ::

        >>> dust = supriya.ugens.Dust.ar(
        ...     density=23,
        ... )
        >>> dust
        Dust.ar()

    """

    _ordered_input_names = collections.OrderedDict([("density", 0.0)])
    _signal_range = SignalRange.UNIPOLAR
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Dust2(UGen):
    """
    A bipolar random impulse generator.

    ::

        >>> dust_2 = supriya.ugens.Dust2.ar(
        ...     density=23,
        ... )
        >>> dust_2
        Dust2.ar()

    """

    _ordered_input_names = collections.OrderedDict([("density", 0.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class ExpRand(UGen):
    """
    An exponential random distribution.

    ::

        >>> exp_rand = supriya.ugens.ExpRand.ir()
        >>> exp_rand
        ExpRand.ir()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict([("minimum", 0.0), ("maximum", 1.0)])
    _valid_calculation_rates = (CalculationRate.SCALAR,)

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, maximum=None, minimum=None):
        minimum, maximum = sorted([minimum, maximum])
        UGen.__init__(
            self, calculation_rate=calculation_rate, minimum=minimum, maximum=maximum
        )


class GrayNoise(UGen):
    """
    A gray noise unit generator.

    ::

        >>> supriya.ugens.GrayNoise.ar()
        GrayNoise.ar()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Hasher(UGen):
    """
    A signal hasher.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> hasher = supriya.ugens.Hasher.ar(
        ...     source=source,
        ... )
        >>> hasher
        Hasher.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", None)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class IRand(UGen):
    """
    An integer uniform random distribution.

    ::

        >>> supriya.ugens.IRand.ir()
        IRand.ir()

    """

    _ordered_input_names = collections.OrderedDict([("minimum", 0), ("maximum", 127)])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class LFClipNoise(UGen):
    """
    A dynamic clipped noise generator.

    ::

        >>> supriya.ugens.LFClipNoise.ar()
        LFClipNoise.ar()

    """

    _ordered_input_names = collections.OrderedDict([("frequency", 500.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFDClipNoise(UGen):
    """
    A clipped noise generator.

    ::

        >>> supriya.ugens.LFDClipNoise.ar()
        LFDClipNoise.ar()

    """

    _ordered_input_names = collections.OrderedDict([("frequency", 500.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFDNoise0(UGen):
    """
    A dynamic step noise generator.

    ::

        >>> supriya.ugens.LFDNoise0.ar()
        LFDNoise0.ar()

    """

    _ordered_input_names = collections.OrderedDict([("frequency", 500.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFDNoise1(UGen):
    """
    A dynamic ramp noise generator.

    ::

        >>> supriya.ugens.LFDNoise1.ar()
        LFDNoise1.ar()

    """

    _ordered_input_names = collections.OrderedDict([("frequency", 500.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFDNoise3(UGen):
    """
    A dynamic polynomial noise generator.

    ::

        >>> supriya.ugens.LFDNoise3.ar()
        LFDNoise3.ar()

    """

    _ordered_input_names = collections.OrderedDict([("frequency", 500.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFNoise0(UGen):
    """
    A step noise generator.

    ::

        >>> supriya.ugens.LFNoise0.ar()
        LFNoise0.ar()

    """

    _ordered_input_names = collections.OrderedDict([("frequency", 500.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFNoise1(UGen):
    """
    A ramp noise generator.

    ::

        >>> supriya.ugens.LFNoise1.ar()
        LFNoise1.ar()

    """

    _ordered_input_names = collections.OrderedDict([("frequency", 500.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LFNoise2(UGen):
    """
    A quadratic noise generator.

    ::

        >>> supriya.ugens.LFNoise2.ar()
        LFNoise2.ar()

    """

    _ordered_input_names = collections.OrderedDict([("frequency", 500.0)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class LinRand(UGen):
    """
    A skewed linear random distribution.

    ::

        >>> lin_rand = supriya.ugens.LinRand.ir(
        ...     minimum=-1.0,
        ...     maximum=1.0,
        ...     skew=0.5,
        ... )
        >>> lin_rand
        LinRand.ir()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0.0), ("maximum", 1.0), ("skew", 0)]
    )
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class Logistic(UGen):
    """
    A chaotic noise function.

    ::

        >>> logistic = supriya.ugens.Logistic.ar(
        ...     chaos_parameter=3.0,
        ...     frequency=1000,
        ...     initial_y=0.5,
        ... )
        >>> logistic
        Logistic.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("chaos_parameter", 3), ("frequency", 1000), ("initial_y", 0.5)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class MantissaMask(UGen):
    """
    A floating-point mantissa mask.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> mantissa_mask = supriya.ugens.MantissaMask.ar(
        ...     source=source,
        ...     bits=3,
        ... )
        >>> mantissa_mask
        MantissaMask.ar()

    """

    _ordered_input_names = collections.OrderedDict([("source", 0), ("bits", 3)])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class NRand(UGen):
    """
    A sum of `n` uniform distributions.

    ::

        >>> n_rand = supriya.ugens.NRand.ir(
        ...     minimum=-1,
        ...     maximum=1,
        ...     n=1,
        ... )
        >>> n_rand
        NRand.ir()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0.0), ("maximum", 1.0), ("n", 1)]
    )
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class PinkNoise(UGen):
    """
    A pink noise unit generator.

    ::

        >>> supriya.ugens.PinkNoise.ar()
        PinkNoise.ar()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class Rand(UGen):
    """
    A uniform random distribution.

    ::

        >>> supriya.ugens.Rand.ir()
        Rand.ir()

    """

    _ordered_input_names = collections.OrderedDict([("minimum", 0.0), ("maximum", 1.0)])
    _valid_calculation_rates = (CalculationRate.SCALAR,)


class RandID(WidthFirstUGen):
    """
    Sets the synth's random generator ID.

    ::

        >>> rand_id = supriya.ugens.RandID.ir(
        ...     rand_id=1,
        ... )
        >>> rand_id
        RandID.ir()

    """

    _ordered_input_names = collections.OrderedDict([("rand_id", 1)])
    _valid_calculation_rates = (CalculationRate.CONTROL, CalculationRate.SCALAR)


class RandSeed(WidthFirstUGen):
    """
    Sets the synth's random generator seed.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> rand_seed = supriya.ugens.RandSeed.ar(
        ...     seed=1,
        ...     trigger=trigger,
        ... )
        >>> rand_seed
        RandSeed.ar()

    """

    _ordered_input_names = collections.OrderedDict([("trigger", 0), ("seed", 56789)])
    _valid_calculation_rates = (
        CalculationRate.AUDIO,
        CalculationRate.CONTROL,
        CalculationRate.SCALAR,
    )


class TExpRand(UGen):
    """
    A triggered exponential random number generator.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> t_exp_rand = supriya.ugens.TExpRand.ar(
        ...     minimum=-1.0,
        ...     maximum=1.0,
        ...     trigger=trigger,
        ... )
        >>> t_exp_rand
        TExpRand.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0.01), ("maximum", 1), ("trigger", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class TIRand(UGen):
    """
    A triggered integer random number generator.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> t_i_rand = supriya.ugens.TIRand.ar(
        ...     minimum=0,
        ...     maximum=127,
        ...     trigger=trigger,
        ... )
        >>> t_i_rand
        TIRand.ar()

    """

    ### CLASS VARIABLES ###

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0), ("maximum", 127), ("trigger", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)

    ### INITIALIZER ###

    def __init__(self, calculation_rate=None, maximum=127, minimum=0, trigger=0):
        minimum = int(minimum)
        maximum = int(maximum)
        UGen.__init__(
            self,
            calculation_rate=calculation_rate,
            maximum=maximum,
            minimum=minimum,
            trigger=trigger,
        )


class TRand(UGen):
    """
    A triggered random number generator.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> t_rand = supriya.ugens.TRand.ar(
        ...     minimum=-1.0,
        ...     maximum=1.0,
        ...     trigger=trigger,
        ... )
        >>> t_rand
        TRand.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("minimum", 0), ("maximum", 1), ("trigger", 0)]
    )
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class TWindex(UGen):
    """
    A triggered windex.

    ::

        >>> trigger = supriya.ugens.Impulse.ar()
        >>> t_windex = supriya.ugens.TWindex.ar(
        ...     trigger=trigger,
        ...     normalize=0,
        ...     array=[1, 2, 3],
        ... )
        >>> t_windex
        TWindex.ar()

    """

    _ordered_input_names = collections.OrderedDict(
        [("trigger", None), ("normalize", 0), ("array", None)]
    )
    _unexpanded_input_names = ("array",)
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)


class WhiteNoise(UGen):
    """
    A white noise unit generator.

    ::

        >>> supriya.ugens.WhiteNoise.ar()
        WhiteNoise.ar()

    """

    _ordered_input_names: UGenInputMap = collections.OrderedDict([])
    _valid_calculation_rates = (CalculationRate.AUDIO, CalculationRate.CONTROL)



import collections
import threading
from typing import Dict, List, Tuple

import uqbar.graphs

import supriya.commands
from supriya.nonrealtime.bases import SessionObject
from supriya.system import SupriyaValueObject
from supriya.utils import iterate_nwise

_local = threading.local()
_local._do_not_propagate_stack = []


class State(SessionObject):
    """
    A non-realtime state.
    """

    ### CLASS VARIABLES ###

    __slots__ = (
        "_transitions",
        "_nodes_to_children",
        "_nodes_to_parents",
        "_offset",
        "_session",
        "_start_buffers",
        "_start_nodes",
        "_stop_buffers",
        "_stop_nodes",
    )

    _ordered_buffer_request_types = (supriya.commands.BufferZeroRequest,)

    ### INITIALIZER ###

    def __init__(self, session, offset):
        from supriya.nonrealtime import Node

        SessionObject.__init__(self, session)
        self._transitions = collections.OrderedDict()
        self._nodes_to_children: Dict[Node, Tuple[Node]] = {}
        self._nodes_to_parents: Dict[Node, Tuple[Node]] = {}
        self._start_nodes = set()
        self._stop_nodes = set()
        self._start_buffers = set()
        self._stop_buffers = set()
        self._offset = offset

    ### SPECIAL METHODS ###

    def __repr__(self):
        return "<{} @{!r}>".format(type(self).__name__, self.offset)

    ### PRIVATE METHODS ###

    @classmethod
    def _apply_transitions(
        cls,
        transitions=None,
        nodes_to_children=None,
        nodes_to_parents=None,
        stop_nodes=None,
    ):
        import supriya.nonrealtime

        if nodes_to_children is not None:
            nodes_to_children = nodes_to_children.copy()
        else:
            nodes_to_children = {}
        if nodes_to_parents is not None:
            nodes_to_parents = nodes_to_parents.copy()
        else:
            nodes_to_parents = {}
        transitions = transitions or {}
        for node, action in transitions.items():
            action.apply_transform(nodes_to_children, nodes_to_parents)
        stop_nodes = stop_nodes or ()
        for stop_node in stop_nodes:
            supriya.nonrealtime.NodeTransition.free_node(
                stop_node, nodes_to_children, nodes_to_parents
            )
        return nodes_to_children, nodes_to_parents

    def _as_graphviz_graph(self):
        from .nodes import Synth

        ordered_synths = []
        cluster = uqbar.graphs.Graph(
            is_cluster=True,
            # attributes={'rank': 'same'},
        )
        node_mapping = {self.session.root_node: cluster}
        for parent, child in self._iterate_node_pairs(
            self.session.root_node, self._nodes_to_children
        ):
            if isinstance(child, Synth) and child not in ordered_synths:
                child_node = child._as_graphviz_node(self.offset)
                ordered_synths.append(child_node)
            else:
                child_node = uqbar.graphs.Graph(
                    is_cluster=True,
                    attributes={
                        "label": child.session_id,
                        "style": ["dashed", "rounded"],
                    },
                )
            node_mapping[child] = child_node
            parent_node = node_mapping[parent]
            parent_node.append(child_node)
        for synth_a, synth_b in iterate_nwise(ordered_synths):
            synth_a.attach(synth_b)
        return cluster, node_mapping, ordered_synths

    def _clone(self, new_offset):
        if float("-inf") < self.offset:
            self.session._apply_transitions(self.offset, chain=False)
        state = type(self)(self.session, new_offset)
        state._nodes_to_children = self.nodes_to_children.copy()
        state._nodes_to_parents = self.nodes_to_parents.copy()
        if new_offset == self.offset:
            state._transitions = self._transitions.copy()
            state._start_buffers.update(self.start_buffers)
            state._stop_buffers.update(self.stop_buffers)
            state._start_nodes.update(self.start_nodes)
            state._stop_nodes.update(self.stop_nodes)
        return state

    def _desparsify(self):
        if self._nodes_to_children is not None:
            return
        previous_state = self.session._find_state_before(
            self.offset, with_node_tree=True
        )
        self._nodes_to_children = previous_state.nodes_to_children.copy()
        self._nodes_to_parents = previous_state.nodes_to_parents.copy()

    def _sparsify(self):
        if self.is_sparse:
            self.session._remove_state_at(self.offset)

    @classmethod
    def _find_first_inconsistency(
        cls, root_node, nodes_to_children_one, nodes_to_children_two, stop_nodes
    ):
        import supriya.nonrealtime

        for parent in cls._iterate_nodes(root_node, nodes_to_children_one):
            if parent in stop_nodes:
                continue
            children_one = nodes_to_children_one.get(parent) or ()
            children_one = [node for node in children_one if node not in stop_nodes]
            children_two = nodes_to_children_two.get(parent) or ()
            if children_one == children_two or not children_two:
                continue
            for i, child in enumerate(children_two):
                if not children_one:
                    action = "ADD_TO_HEAD"
                    target = parent
                elif len(children_one) <= i:
                    action = "ADD_AFTER"
                    target = children_one[i - 1]
                elif children_one[i] is not child:
                    action = "ADD_BEFORE"
                    target = children_one[i]
                else:
                    continue
                transition = supriya.nonrealtime.NodeTransition(
                    source=child, target=target, action=action
                )
                return transition

    @classmethod
    def _iterate_nodes(cls, root_node, nodes_to_children):
        def recurse(parent):
            yield parent
            children = nodes_to_children.get(parent, ()) or ()
            for child in children:
                for node in recurse(child):
                    yield node

        return recurse(root_node)

    @classmethod
    def _iterate_node_pairs(cls, root_node, nodes_to_children):
        def recurse(parent):
            children = nodes_to_children.get(parent, ()) or ()
            for child in children:
                yield parent, child
                for pair in recurse(child):
                    yield pair

        return recurse(root_node)

    @classmethod
    def _rebuild_transitions(cls, state_one, state_two):
        # print('REBUILDING')
        assert state_one.session.root_node is state_two.session.root_node
        a_children = state_one.nodes_to_children.copy()
        a_parents = state_one.nodes_to_parents.copy()
        b_children, b_parents = a_children.copy(), a_parents.copy()
        stop_nodes = state_two.stop_nodes
        transitions = collections.OrderedDict()
        counter = 0
        while b_children != state_two.nodes_to_children:
            # print('ROUND', counter)
            # print('C-1', b_children)
            # print('C-2', state_two.nodes_to_children)
            transition = State._find_first_inconsistency(
                state_one.session.root_node,
                b_children,
                state_two.nodes_to_children,
                stop_nodes,
            )
            if transition is not None:
                transitions[transition.source] = transition
            b_children, b_parents = State._apply_transitions(
                transitions, a_children, a_parents, stop_nodes
            )
            counter += 1
            if counter == 100:
                raise Exception
        return transitions

    ### PUBLIC METHODS ###

    def report(self):
        state = {}
        node_hierarchy = {}
        items = sorted(
            self.nodes_to_children.items(), key=lambda item: item[0].session_id
        )
        for parent, children in items:
            if not children:
                children = []
            node_hierarchy[str(parent)] = [str(child) for child in children]
        node_lifecycle = {}
        if self.start_nodes:
            node_lifecycle["start"] = sorted(str(node) for node in self.start_nodes)
        if self.stop_nodes:
            node_lifecycle["stop"] = sorted(str(node) for node in self.stop_nodes)
        if node_hierarchy:
            state["hierarchy"] = node_hierarchy
        if node_lifecycle:
            state["lifecycle"] = node_lifecycle
        state["offset"] = self.offset
        return state

    ### PUBLIC PROPERTIES ###

    @property
    def is_sparse(self):
        if self.start_nodes:
            return False
        elif self.stop_nodes:
            return False
        elif self.transitions:
            return False
        return True

    @property
    def nodes_to_children(
        self,
    ) -> Dict["supriya.nonrealtime.Node", Tuple["supriya.nonrealtime.Node"]]:
        return self._nodes_to_children

    @property
    def nodes_to_parents(
        self,
    ) -> Dict["supriya.nonrealtime.Node", Tuple["supriya.nonrealtime.Node"]]:
        return self._nodes_to_parents

    @property
    def offset(self) -> float:
        return self._offset

    @property
    def start_buffers(self):
        return self._start_buffers

    @property
    def start_nodes(self):
        return self._start_nodes

    @property
    def stop_buffers(self):
        return self._stop_buffers

    @property
    def stop_nodes(self):
        return self._stop_nodes

    @property
    def overlap_nodes(self):
        interval_tree = self.session._nodes
        intersection = interval_tree.find_intersection(self.offset)
        overlap = [_ for _ in intersection if _.start_offset < self.offset]
        return overlap

    @property
    def overlap_buffers(self):
        interval_tree = self.session._buffers
        intersection = interval_tree.find_intersection(self.offset)
        overlap = [_ for _ in intersection if _.start_offset < self.offset]
        return overlap

    @property
    def transitions(self):
        return self._transitions


class Moment(SessionObject):
    """
    A moment-in-time referencing a singleton non-realtime state.

    ::

        >>> import supriya.nonrealtime
        >>> session = supriya.nonrealtime.Session()
        >>> moment = session.at(10.5)

    """

    ### CLASS VARIABLES ###

    __slots__ = ("_offset", "_propagate", "_session", "_state")

    ### INITIALIZER ###

    def __init__(self, session, offset, state, propagate=True):
        SessionObject.__init__(self, session)
        self._offset = offset
        self._state = state
        self._propagate = bool(propagate)

    ### SPECIAL METHODS ###

    def __enter__(self):
        self.session.active_moments.append(self)
        if self.propagate:
            self.session._apply_transitions(self.state.offset)
        return self

    def __eq__(self, expr):
        if not isinstance(expr, type(self)):
            return False
        if expr.session is not self.session:
            return False
        return expr.offset == self.offset

    def __exit__(self, exc_type, exc_value, traceback):
        self.session.active_moments.pop()
        if self.propagate:
            self.session._apply_transitions(self.state.offset)

    def __lt__(self, expr):
        if not isinstance(expr, type(self)) or expr.session is not self.session:
            raise ValueError(expr)
        return self.offset < expr.offset

    def __repr__(self):
        return "<{} @{!r}>".format(type(self).__name__, self.offset)

    ### PUBLIC PROPERTIES ###

    @property
    def offset(self):
        return self._offset

    @property
    def propagate(self):
        return self._propagate

    @property
    def state(self):
        return self._state


class NodeTransition(SupriyaValueObject):
    """
    A non-realtime state transition.
    """

    ### CLASS VARIABLES ###

    __slots__ = ("_source", "_target", "_action")

    ### INITIALIZER ###

    def __init__(self, source=None, action=None, target=None):
        if action is not None:
            action = supriya.AddAction.from_expr(action)
            assert isinstance(action, supriya.AddAction)
            assert source is not target
        if action is None:
            assert source is not None
            assert target is None
        self._action = action
        self._source = source
        self._target = target

    ### PRIVATE METHODS ###

    def _free_node(self, nodes_to_children, nodes_to_parents):
        node = self.source
        for child in nodes_to_children.get(node, ()) or ():
            self.free_node(child, nodes_to_children, nodes_to_parents)
        parent = nodes_to_parents.get(node, None)
        if node in nodes_to_children:
            del nodes_to_children[node]
        if node in nodes_to_parents:
            del nodes_to_parents[node]
        if not parent:
            return
        children = list(nodes_to_children[parent])
        children.remove(node)
        nodes_to_children[parent] = tuple(children) or None

    def _move_node(self, nodes_to_children, nodes_to_parents):
        assert self.target in nodes_to_children
        if self.source not in nodes_to_children:
            nodes_to_children[self.source] = None
        old_parent = nodes_to_parents.get(self.source, None)
        if old_parent:
            children = list(nodes_to_children[old_parent])
            children.remove(self.source)
            nodes_to_children[old_parent] = tuple(children) or None
        if self.action in (supriya.AddAction.ADD_AFTER, supriya.AddAction.ADD_BEFORE):
            new_parent = nodes_to_parents[self.target]
        else:
            new_parent = self.target
        nodes_to_parents[self.source] = new_parent
        children = list(nodes_to_children.get(new_parent, None) or ())
        if self.action == supriya.AddAction.ADD_TO_HEAD:
            children.insert(0, self.source)
        elif self.action == supriya.AddAction.ADD_TO_TAIL:
            children.append(self.source)
        elif self.action == supriya.AddAction.ADD_BEFORE:
            index = children.index(self.target)
            children.insert(index, self.source)
        elif self.action == supriya.AddAction.ADD_AFTER:
            index = children.index(self.target) + 1
            children.insert(index, self.source)
        nodes_to_children[new_parent] = tuple(children) or None

    def _to_request(self, id_mapping):
        node_id_pair = (id_mapping[self.source], id_mapping[self.target])
        if self.action == supriya.AddAction.ADD_TO_HEAD:
            request_class = supriya.commands.GroupHeadRequest
        elif self.action == supriya.AddAction.ADD_TO_TAIL:
            request_class = supriya.commands.GroupTailRequest
        elif self.action == supriya.AddAction.ADD_BEFORE:
            request_class = supriya.commands.NodeBeforeRequest
        elif self.action == supriya.AddAction.ADD_AFTER:
            request_class = supriya.commands.NodeAfterRequest
        request = request_class(node_id_pairs=[node_id_pair])
        return request

    ### PUBLIC METHODS ###

    def apply_transform(self, nodes_to_children, nodes_to_parents):
        if self.action is None:
            self._free_node(nodes_to_children, nodes_to_parents)
        else:
            self._move_node(nodes_to_children, nodes_to_parents)

    @classmethod
    def free_node(cls, node, nodes_to_children, nodes_to_parents):
        action = cls(source=node)
        action.apply_transform(nodes_to_children, nodes_to_parents)

    ### PUBLIC PROPERTIES ###

    @property
    def action(self):
        return self._action

    @property
    def source(self):
        return self._source

    @property
    def target(self):
        return self._target


class DoNotPropagate:
    """
    Context manager which prevents propagation of node hierarchy changes across
    states.
    """

    ### CLASS VARIABLES ###

    _stack: List["DoNotPropagate"] = _local._do_not_propagate_stack

    ### SPECIAL METHODS ###

    def __enter__(self):
        self._stack.append(self)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._stack.pop()
